"""
### Utility functions to work with GDAL and projections ###
"""

# Standard Library
import sys; sys.path.append("../../")
from typing import Union, List

# External
import osgeo
from osgeo import gdal, ogr, osr
import numpy as np

# Internal
from buteo.utils import utils_gdal


def _get_default_projection() -> str:
    """
    Get the default projection for a new raster.
    EPSG:4326 in WKT format.

    Returns
    -------
    str:
        The default projection. (EPSG:4326) in WKT format.
    """
    epsg_4326_wkt = 'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]'

    return epsg_4326_wkt


def _get_default_projection_osr() -> osr.SpatialReference:
    """
    Get the default projection for a new raster.
    EPSG:4326 in osr.SpatialReference format.

    Returns
    -------
    osr.SpatialReference:
        The default projection. (EPSG:4326) in osr.SpatialReference format.
    """
    epsg_4326_wkt = _get_default_projection()

    spatial_ref = osr.SpatialReference()
    spatial_ref.ImportFromWkt(epsg_4326_wkt)

    return spatial_ref


def _get_esri_projection(esri_code: str) -> str:
    """
    Imports a projection from an ESRI code.

    Parameters
    ----------
    esri_code : str
        The ESRI code to import.

    Returns
    -------
    str:
        The projection in WKT format.

    Notes
    -----
    The following ESRI codes are currently supported:
    - ESRI:54009 (EPSG:3785) ~ World_Mollweide
    """

    # ESRI:54009 is equivalent to EPSG:3785
    if esri_code == "ESRI:54009":
        wkt_code = """
        PROJCRS["World_Mollweide",
            BASEGEOGCRS["WGS 84",
                DATUM["World Geodetic System 1984",
                    ELLIPSOID["WGS 84",6378137,298.257223563,
                        LENGTHUNIT["metre",1]]],
                PRIMEM["Greenwich",0,
                    ANGLEUNIT["Degree",0.0174532925199433]]],
            CONVERSION["World_Mollweide",
                METHOD["Mollweide"],
                PARAMETER["Longitude of natural origin",0,
                    ANGLEUNIT["Degree",0.0174532925199433],
                    ID["EPSG",8802]],
                PARAMETER["False easting",0,
                    LENGTHUNIT["metre",1],
                    ID["EPSG",8806]],
                PARAMETER["False northing",0,
                    LENGTHUNIT["metre",1],
                    ID["EPSG",8807]]],
            CS[Cartesian,2],
                AXIS["(E)",east,
                    ORDER[1],
                    LENGTHUNIT["metre",1]],
                AXIS["(N)",north,
                    ORDER[2],
                    LENGTHUNIT["metre",1]],
            USAGE[
                SCOPE["Not known."],
                AREA["World."],
                BBOX[-90,-180,90,180]],
            ID["ESRI",54009]]
        """
    else:
        raise ValueError("Unknown ESRI code")

    # Create a spatial reference object
    spatial_ref = osr.SpatialReference()

    # Import the EPSG code
    spatial_ref.ImportFromWkt(wkt_code)

    return spatial_ref.ExportToWkt()


def parse_projection(
    projection: Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference],
    return_wkt: bool = False,
) -> Union[osr.SpatialReference, str]:
    """
    Parses a gdal, ogr or osr data source and extracts the projection. If
    a string or int is passed, it attempts to open it and return the projection as
    an osr.SpatialReference.

    Parameters
    ----------
    projection : Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]
        The projection to parse.

    return_wkt : bool, optional
        Whether to return the projection as a WKT string or an osr.SpatialReference

    Returns
    -------
    Union[osr.SpatialReference, str]:
        The projection as an osr.SpatialReference or a WKT string.
    """
    valid_types = (str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference)
    if not isinstance(projection, valid_types):
        raise ValueError("projection must be a string, int, gdal.Dataset, ogr.DataSource, or osr.SpatialReference.")

    err_msg = f"Unable to parse target projection: {projection}"
    target_proj = osr.SpatialReference()

    # Suppress gdal errors and handle them ourselves.
    gdal.PushErrorHandler("CPLQuietErrorHandler")

    try:
        if isinstance(projection, ogr.DataSource):
            layer = projection.GetLayer()
            target_proj = layer.GetSpatialRef()

        elif isinstance(projection, gdal.Dataset):
            target_proj.ImportFromWkt(projection.GetProjection())

        elif isinstance(projection, osr.SpatialReference):
            if projection.ExportToWkt() == "":
                raise ValueError("Spatial reference is empty.")
            target_proj = projection

        elif isinstance(projection, str):
            if utils_gdal._check_is_raster(projection):
                ref = gdal.Open(projection, 0)
                target_proj.ImportFromWkt(ref.GetProjection())
            elif utils_gdal._check_is_vector(projection):
                ref = ogr.Open(projection, 0)
                layer = ref.GetLayer()
                target_proj = layer.GetSpatialRef()
            else:
                if projection.lower().startswith("epsg:"):
                    try:
                        success = target_proj.ImportFromEPSG(int(projection.split(":")[1])) == 0
                    except: # pylint: disable=bare-except
                        success = False
                elif projection.lower().startswith("esri:"):
                    try:
                        success = target_proj.ImportFromWkt(_get_esri_projection(projection)) == 0
                    except: # pylint: disable=bare-except
                        success = False
                else:
                    try:
                        success = target_proj.ImportFromWkt(projection) == 0
                    except: # pylint: disable=bare-except
                        success = False
                        try:
                            success = target_proj.ImportFromProj4(projection) == 0
                        except: # pylint: disable=bare-except
                            success = False

                if not success:
                    raise ValueError(err_msg)

        elif isinstance(projection, int):
            if target_proj.ImportFromEPSG(projection) != 0:
                raise ValueError(err_msg)

        else:
            raise ValueError(err_msg)

    finally:
        gdal.PopErrorHandler()

    if not isinstance(target_proj, osr.SpatialReference) or target_proj.ExportToWkt() == "":
        raise ValueError(err_msg)

    return target_proj.ExportToWkt() if return_wkt else target_proj


def _projection_is_latlng(projection: Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]) -> bool:
    """
    Tests if a projection is in latlng format.

    Parameters
    ----------
    projection : Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]
        The projection to test.

    Returns
    -------
    bool:
        **True** if the projection is in latlng format, **False** otherwise.
    """
    proj = parse_projection(projection)

    if proj.IsGeographic():
        return True

    return False


def _check_projections_match(
    source1: Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference],
    source2: Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference],
) -> bool:
    """
    Tests if two projection sources have the same projection.

    Parameters
    ----------
    source1 : Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]
        The first projection to test.

    source2 : Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]
        The second projection to test.

    Returns
    -------
    bool:
        **True** if the projections match, **False** otherwise.

    """
    proj1 = parse_projection(source1)
    proj2 = parse_projection(source2)

    if proj1.IsSame(proj2):
        return True
    elif proj1.ExportToProj4() == proj2.ExportToProj4():
        return True

    return False


def _check_projections_match_list(
    list_of_projection_sources: List[Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]],
) -> bool:
    """
    Tests if a list of projection sources all have the same projection.

    Parameters
    ----------
    list_of_projection_sources : List[Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]]
        The list of projections to test.

    Returns
    -------
    bool:
        **True** if the projections match, **False** otherwise.
    """
    assert isinstance(list_of_projection_sources, list), "list_of_projection_sources must be a list."

    if len(list_of_projection_sources) == 0:
        raise ValueError("list_of_projection_sources must not be empty.")

    if len(list_of_projection_sources) == 1:
        return True

    first = None
    for index, source in enumerate(list_of_projection_sources):
        if index == 0:
            first = parse_projection(source)
        else:
            compare = parse_projection(source)

            if not first.IsSame(compare) and first.ExportToProj4() != compare.ExportToProj4():
                return False

    return True



def _get_projection_from_raster(
    raster: Union[str, gdal.Dataset],
) -> osr.SpatialReference:
    """
    Get the projection from a raster.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset]
        A raster or raster path.
    
    Returns
    -------
    osr.SpatialReference:
        The projection in OSR format.
    """
    opened = None
    if isinstance(raster, gdal.Dataset):
        opened = raster
    else:
        try:
            opened = gdal.Open(raster, gdal.GA_ReadOnly)
        except ValueError:
            opened = None

    if opened is None:
        raise RuntimeError(f"Could not open raster. {raster}")

    projection = osr.SpatialReference()
    projection.ImportFromWkt(opened.GetProjection())
    opened = None

    return projection


def _get_projection_from_vector(
    vector: Union[str, ogr.DataSource],
) -> osr.SpatialReference:
    """
    Get the projection from a vector.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource]
        A vector or vector path.

    Returns
    -------
    osr.SpatialReference:
        The projection in OSR format.
    """
    opened = None
    if isinstance(vector, ogr.DataSource):
        opened = vector
    else:
        try:
            opened = ogr.Open(vector, gdal.GA_ReadOnly)
        except ValueError:
            opened = None

    if opened is None:
        raise RuntimeError(f"Could not open vector. {vector}")

    layer = opened.GetLayer()
    projection = layer.GetSpatialRef()
    opened = None

    return projection


def _get_projection_from_dataset(
    dataset: Union[str, gdal.Dataset, ogr.DataSource],
) -> osr.SpatialReference:
    """
    Get the projection from a dataset.

    Parameters
    ----------
    dataset : Union[str, gdal.Dataset, ogr.DataSource]
        A dataset or dataset path.
    
    Returns
    -------
    osr.SpatialReference:
        The projection in OSR format.
    """
    assert isinstance(dataset, (str, gdal.Dataset, ogr.DataSource)), "DataSet must be a string, ogr.DataSource, or gdal.Dataset."

    opened = dataset if isinstance(dataset, (gdal.Dataset, ogr.DataSource)) else None

    if opened is None:
        try:
            opened = gdal.Open(dataset, gdal.GA_ReadOnly)
        except ValueError:
            try:
                opened = ogr.Open(dataset, gdal.GA_ReadOnly)
            except ValueError:
                opened = None

    if opened is None:
        raise RuntimeError(f"Could not open dataset. {dataset}")

    if isinstance(opened, gdal.Dataset):
        return _get_projection_from_raster(opened)

    if isinstance(opened, ogr.DataSource):
        return _get_projection_from_vector(opened)

    raise RuntimeError(f"Could not get projection from dataset. {dataset}")


def reproject_bbox(
    bbox_ogr: List[Union[int, float]],
    source_projection: Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference],
    target_projection: Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference],
) -> List[Union[int, float]]:
    """
    Reprojects an OGR formatted bbox.
    OGR formatted bboxes are in the format `[x_min, x_max, y_min, y_max]`.

    Parameters
    ----------
    bbox_ogr : List[Union[int, float]]
        The OGR formatted bbox to reproject.

    source_projection : Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]
        The source projection.

    target_projection : Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]
        The target projection.

    Returns
    -------
    List[Union[int, float]]:
        The reprojected bbox.
    """
    assert isinstance(bbox_ogr, list), "bbox_ogr must be a list."
    assert len(bbox_ogr) == 4, "bbox_ogr must have 4 elements."
    for val in bbox_ogr:
        assert isinstance(val, (int, float)), "bbox_ogr must only contain numbers."
    assert isinstance(source_projection, (str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference)), "source_projection must be a string, int, gdal.Dataset, ogr.DataSource, or osr.SpatialReference."
    assert isinstance(target_projection, (str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference)), "target_projection must be a string, int, gdal.Dataset, ogr.DataSource, or osr.SpatialReference."

    x_min, x_max, y_min, y_max = bbox_ogr
    assert x_min <= x_max, "x_min must be less than or equal to x_max."
    assert y_min <= y_max, "y_min must be less than or equal to y_max."

    src_proj = parse_projection(source_projection)
    dst_proj = parse_projection(target_projection)

    if _check_projections_match(src_proj, dst_proj):
        return bbox_ogr


    p1 = [x_min, y_min]
    p2 = [x_max, y_min]
    p3 = [x_max, y_max]
    p4 = [x_min, y_max]

    options = osr.CoordinateTransformationOptions()
    if _projection_is_latlng(src_proj):
        p1.reverse()
        p2.reverse()
        p3.reverse()
        p4.reverse()

        options.SetAreaOfInterest(-180.0, -90.0, 180.0, 90.0)

    transformer = osr.CoordinateTransformation(src_proj, dst_proj, options)

    gdal.PushErrorHandler("CPLQuietErrorHandler")
    try:
        p1t = transformer.TransformPoint(p1[0], p1[1])
        p2t = transformer.TransformPoint(p2[0], p2[1])
        p3t = transformer.TransformPoint(p3[0], p3[1])
        p4t = transformer.TransformPoint(p4[0], p4[1])
    except RuntimeError:
        p1t = transformer.TransformPoint(float(p1[0]), float(p1[1]))
        p2t = transformer.TransformPoint(float(p2[0]), float(p2[1]))
        p3t = transformer.TransformPoint(float(p3[0]), float(p3[1]))
        p4t = transformer.TransformPoint(float(p4[0]), float(p4[1]))
    gdal.PopErrorHandler()

    transformed_x_min = min(p1t[0], p2t[0], p3t[0], p4t[0])
    transformed_x_max = max(p1t[0], p2t[0], p3t[0], p4t[0])
    transformed_y_min = min(p1t[1], p2t[1], p3t[1], p4t[1])
    transformed_y_max = max(p1t[1], p2t[1], p3t[1], p4t[1])

    return [
        transformed_x_min,
        transformed_x_max,
        transformed_y_min,
        transformed_y_max,
    ]


def _reproject_point(
    p: List[Union[int, float]],
    source_projection: Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference],
    target_projection: Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference],
) -> List[Union[int, float]]:
    """
    Reprojects a point.

    Parameters
    ----------
    p : List[Union[int, float]]
        The point to reproject.

    source_projection : Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]
        The source projection.

    target_projection : Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]
        The target projection.

    Returns
    -------
    List[Union[int, float]]:
        The reprojected point.
    """
    assert isinstance(p, list), "p must be a list."
    assert len(p) == 2, "p must have 2 elements."

    src_proj = parse_projection(source_projection)
    dst_proj = parse_projection(target_projection)

    if _check_projections_match(src_proj, dst_proj):
        return p

    options = osr.CoordinateTransformationOptions()

    if _projection_is_latlng(dst_proj):
        options.SetAreaOfInterest(-180.0, -90.0, 180.0, 90.0)

    transformer = osr.CoordinateTransformation(src_proj, dst_proj, options)

    gdal.PushErrorHandler("CPLQuietErrorHandler")
    try:
        pt = transformer.TransformPoint(p[0], p[1])
    except RuntimeError:
        pt = transformer.TransformPoint(float(p[0]), float(p[1]))
    gdal.PopErrorHandler()

    return [pt[0], pt[1]]


def _get_utm_zone_from_latlng(
    latlng: List[Union[int, float]],
    return_epsg: bool = False,
) -> str:
    """
    Get the UTM ZONE from a latlng list.

    Parameters
    ----------
    latlng : List[Union[int, float]]
        The latlng list to get the UTM ZONE from.

    return_epsg : bool, optional
        Whether or not to return the EPSG code instead of the WKT, by default False

    Returns
    -------
    str
        The WKT or EPSG code.
    """
    assert isinstance(latlng, (list, np.ndarray)), "latlng must be in the form of a list."

    zone = round(((latlng[1] + 180) / 6) + 1)
    n_or_s = "S" if latlng[0] < 0 else "N"

    false_northing = "10000000" if n_or_s == "S" else "0"
    central_meridian = str(round(((zone * 6) - 180) - 3))
    epsg = f"32{'7' if n_or_s == 'S' else '6'}{str(zone)}"

    if return_epsg:
        return epsg

    wkt = f"""
        PROJCS["WGS 84 / UTM zone {str(zone)}{n_or_s}",
        GEOGCS["WGS 84",
            DATUM["WGS_1984",
                SPHEROID["WGS 84",6378137,298.257223563,
                    AUTHORITY["EPSG","7030"]],
                AUTHORITY["EPSG","6326"]],
            PRIMEM["Greenwich",0,
                AUTHORITY["EPSG","8901"]],
            UNIT["degree",0.0174532925199433,
                AUTHORITY["EPSG","9122"]],
            AUTHORITY["EPSG","4326"]],
        PROJECTION["Transverse_Mercator"],
        PARAMETER["latitude_of_origin",0],
        PARAMETER["central_meridian",{central_meridian}],
        PARAMETER["scale_factor",0.9996],
        PARAMETER["false_easting",500000],
        PARAMETER["false_northing",{false_northing}],
        UNIT["metre",1,
            AUTHORITY["EPSG","9001"]],
        AXIS["Easting",EAST],
        AXIS["Northing",NORTH],
        AUTHORITY["EPSG","{epsg}"]]
    """
    projection = osr.SpatialReference()
    projection.ImportFromWkt(wkt)

    return projection


def _reproject_latlng_point_to_utm(
    latlng: List[Union[int, float]],
) -> List[Union[int, float]]:
    """
    Converts a latlng point into an UTM point.
    Takes point in [lat, lng], returns [utm_x, utm_y].

    Parameters
    ----------
    latlng : List[Union[int, float]]
        The latlng point to convert.

    Returns
    -------
    List[Union[int, float]]
        The converted UTM point.
    """
    source_projection = osr.SpatialReference()
    source_projection_wkt = _get_default_projection()
    source_projection.ImportFromWkt(source_projection_wkt)
    target_projection = _get_utm_zone_from_latlng(latlng)

    transformer = osr.CoordinateTransformation(
        source_projection, target_projection
    )

    try:
        utm_x, utm_y, _utm_z = transformer.TransformPoint(latlng[0], latlng[1])
    except: # pylint: disable=bare-except
        utm_x, utm_y, _utm_z = transformer.TransformPoint(float(latlng[0]), float(latlng[1]))

    return [utm_x, utm_y]



"""
### Generic utility functions ###

Functions that make interacting with the toolbox easier.
"""

# Standard Library
import time
from typing import Any, Union, List, Tuple



def _get_variable_as_list(
    variable_or_list: Union[List, Any],
) -> List[Any]:
    """
    Ensures that a variable is a list. If the variable is a list, return the list.
    If the variable is not a list, return a list with the variable as the only element.

    Parameters
    ----------
    variable_or_list : Union[List, Any]
        The variable to check.

    Returns
    -------
    List[Any]
        The variable as a list.
    """
    if isinstance(variable_or_list, list):
        return variable_or_list

    return [variable_or_list]


def _get_unix_seconds_as_str() -> str:
    """
    Get a string of the current UNIX time in seconds.

    Returns
    -------
    str
        A string of the current UNIX time in seconds.
    """
    return str(int(time.time()))


def _get_time_as_str() -> str:
    """
    Gets the current time as a string.
    in the format: YYYYMMDD_HHMMSS

    Returns
    -------
    str
        The current time as a string.
    """
    return time.strftime("%Y%m%d_%H%M%S")


def _check_variable_is_float(variable: Any) -> bool:
    """
    Check if a variable is a float.
    If it is a string, see if it a representation of a float.

    Parameters
    ----------
    variable : Any
        The variable to check.

    Returns
    -------
    bool
        True if the variable is a float or float representation, False otherwise.
    """
    if isinstance(variable, float):
        return True

    if isinstance(variable, str):
        try:
            float(variable)
            return True

        except ValueError:
            return False

    return False


def _check_variable_is_int(variable: Any) -> bool:
    """
    Check if a variable is an integer.
    If it is a string, see if it a representation of an integer.
    If it is a float, return False.

    Parameters
    ----------
    variable : Any
        The variable to check.

    Returns
    -------
    bool
        True if the variable is an integer or integer representation, False otherwise.
    """
    if isinstance(variable, int):
        return True

    if isinstance(variable, str):
        try:
            int(variable)
            return True
        except ValueError:
            return False

    return False


def _check_variable_is_number_type(variable: Any) -> bool:
    """
    Check if variable is a number.

    Parameters
    ----------
    variable : Any
        The variable to check.

    Returns
    -------
    bool
        True if the variable is a number type, False otherwise.
    """
    if isinstance(variable, (float, int)):
        return True

    return False


def _get_variable_as_number(
    variable: Any,
) -> float:
    """
    Attempts to convert a variable to a number.

    Parameters
    ----------
    variable : Any
        The variable to convert.

    Returns
    -------
    float
        The variable as a float.
    """
    assert isinstance(variable, (str, int, float)), "value must be a string, integer or float."
    if _check_variable_is_int(variable):
        return int(variable)

    if _check_variable_is_float(variable):
        return float(variable)

    raise RuntimeError(f"Could not convert {variable} to a number.")


def _ensure_negative(number: Union[int, float]) -> Union[int, float]:
    """
    Ensures that a valid is negative. If the number is positive, it is made negative.

    Parameters
    ----------
    number : int, float
        A number.

    Returns
    -------
    int, float
        The number, made negative if it was positive.
    """
    assert _check_variable_is_number_type(number), f"number must be a number. Received: {number}"

    if number <= 0:
        return number

    return -number


def _check_variable_is_iterable_or_type(potential_type: Any) -> bool:
    """
    Recursively check if a variable is a type, list, or tuple.

    Parameters
    ----------
    potential_type : Any
        The variable to check.

    Returns
    -------
    bool
        True if the variable is a type, list, or tuple, False otherwise.
    """
    if isinstance(potential_type, type(None)):
        return True

    if isinstance(potential_type, type):
        return True

    if isinstance(potential_type, (list, tuple)):
        for item in potential_type:
            if not _check_variable_is_iterable_or_type(item):
                return False
        return True

    return False


def _type_check(
    variable: Any,
    types: Tuple[type, ...],
    name: str = "",
    *,
    throw_error: bool = True,
) -> bool:
    """
    Utility function to type check the inputs of a function. Checks recursively.
    This is useful for the external facing functions of a module.

    Use like this:
    ```python
    >>> def my_function(
    ...     my_variable: Optional[Union[str, float, List[str]]] = None
    ... ) -> None:
    ...     type_check(my_variable, [str, float, [str], None], "my_variable")
    ...     # Do stuff
    ```

    Parameters
    ----------
    variable : Any
        The variable to check.
    
    types : Tuple[type, ...]
        The types to check against.

    name : str, optional
        The name of the variable to check.

    throw_error : bool, optional
        If True, throw an error if the type check fails. If False, return False if the type check fails.
    
    Returns
    -------
    bool
        True if the type check passes, False otherwise.
    """
    assert isinstance(name, str), "name must be a string."
    assert _check_variable_is_iterable_or_type(types), f"types must be a type, list, None, or tuple. not: {types}"

    if not isinstance(types, (list, tuple)):
        types = [types]

    valid_types = []
    for valid_type in types:
        if valid_type is None:
            valid_types.append(type(None))
        elif isinstance(valid_type, type):
            valid_types.append(valid_type)
        elif isinstance(valid_type, (list, tuple)):
            valid_types.append(valid_type)
        else:
            raise ValueError(f"Invalid type: {valid_type}")

    if not isinstance(variable, (list, tuple)):
        sublist_valid_types = []
        for valid_type in valid_types:
            if not isinstance(valid_type, (list, tuple)):
                sublist_valid_types.append(valid_type)

        for valid_type in sublist_valid_types:
            if isinstance(variable, valid_type):
                return True

    if type(variable) in valid_types:
        return True

    type_list = [type(val) for val in valid_types]

    if isinstance(variable, list) and type([]) in type_list:
        for sublist in valid_types:
            if not isinstance(sublist, list):
                continue

            if len(sublist) == 0:
                return True

            found = 0
            for item in variable:
                if type(item) in sublist:
                    found += 1

            if found == len(variable):
                return True

    if isinstance(variable, tuple) and type(()) in type_list:
        for sublist in valid_types:
            if not isinstance(sublist, tuple):
                continue

            if len(sublist) == 0:
                return True

            found = 0
            for item in variable:
                if type(item) in sublist:
                    found += 1

            if found == len(variable):
                return True
    if throw_error:
        raise ValueError(
            f"The type of variable {name} is not valid. Expected: {types}, got: {type(variable)}"
        )

    return False


def _check_number_is_within_threshold(number, target, threshold):
    """
    Check if a number is within a threshold of a target.

    Parameters
    ----------
    number : float
        The number to check.

    target : float
        The target number.

    threshold : float
        The threshold.

    Returns
    -------
    bool
        True if the number is within the threshold of the target, False otherwise.
    """
    assert isinstance(number, (int, float)), "number must be a number."
    assert isinstance(target, (int, float)), "target must be a number."
    assert isinstance(threshold, (int, float)), "threshold must be a number."

    return abs(number - target) <= threshold


def _check_number_is_within_range(number, min_value, max_value):
    """
    Check if a number is within a range.

    Parameters
    ----------
    number : float
        The number to check.

    min_value : float
        The minimum value.

    max_value : float
        The maximum value.

    Returns
    -------
    bool
        True if the number is within the range, False otherwise.
    """
    assert isinstance(number, (int, float)), "number must be a number."
    assert isinstance(min_value, (int, float)), "min_value must be a number."
    assert isinstance(max_value, (int, float)), "max_value must be a number."

    return min_value <= number <= max_value



"""
### Clip vectors to other geometries ###

Clip vector files with other geometries. Can come from rasters or vectors.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List
from warnings import warn

# External
from osgeo import ogr, gdal, osr

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_gdal,
    utils_path,
    utils_projection,
)
from buteo.vector import core_vector
from buteo.raster import core_raster
from buteo.vector.reproject import _vector_reproject



def _vector_clip(
    vector: Union[str, ogr.DataSource],
    clip_geom: Union[str, ogr.DataSource],
    out_path: Optional[str] = None,
    to_extent: bool = False,
    target_projection: Optional[Union[str, int, ogr.DataSource, gdal.Dataset, osr.SpatialReference]] = None,
    preserve_fid: bool = True,
    promote_to_multi: bool = True,
    overwrite: bool = True,
    verbose: bool = False,
) -> str:
    """ Internal. """
    assert isinstance(vector, (str, ogr.DataSource)), "Invalid vector input."
    assert isinstance(clip_geom, (str, ogr.DataSource)), "Invalid clip_geom input."

    if out_path is None:
        out_path = utils_path._get_temp_filepath(vector, suffix="_clipped", ext="gpkg")
    else:
        assert utils_path._check_is_valid_output_filepath(out_path, overwrite), "Invalid vector output path."

    gdal.UseExceptions()

    if not verbose:
        gdal.PushErrorHandler("CPLQuietErrorHandler")

    input_path = utils_gdal._get_path_from_dataset(vector)

    options = []
    to_clear = []
    geometry_to_clip = None
    if utils_gdal._check_is_raster(clip_geom):
        geometry_to_clip = core_raster.raster_to_extent(clip_geom)
        to_clear.append(geometry_to_clip)

    if utils_gdal._check_is_vector(clip_geom):
        if to_extent:
            geometry_to_clip = core_vector.vector_to_extent(clip_geom)
            to_clear.append(geometry_to_clip)
        else:
            geometry_to_clip = utils_gdal._get_path_from_dataset(clip_geom)
    else:
        raise ValueError(f"Invalid input in clip_geom, unable to parse: {clip_geom}")

    clip_vector_reprojected = _vector_reproject(geometry_to_clip, vector)

    if clip_vector_reprojected != geometry_to_clip:
        to_clear.append(clip_vector_reprojected)

    x_min, x_max, y_min, y_max = core_vector._get_basic_metadata_vector(clip_vector_reprojected)["bbox"]

    options.append(f"-spat {x_min} {y_min} {x_max} {y_max}")
    options.append(f'-clipsrc "{clip_vector_reprojected}"')

    if promote_to_multi:
        options.append("-nlt PROMOTE_TO_MULTI")

    if preserve_fid:
        options.append("-preserve_fid")
    else:
        options.append("-unsetFid")

    if target_projection is not None:
        wkt = utils_projection.parse_projection(target_projection, return_wkt=True).replace(" ", "\\")

        options.append(f'-t_srs "{wkt}"')

    # dst  # src
    success = gdal.VectorTranslate(
        out_path,
        input_path,
        format=utils_gdal._get_vector_driver_name_from_path(out_path),
        options=" ".join(options),
    )

    utils_gdal.delete_dataset_if_in_memory_list(to_clear)

    if not verbose:
        gdal.PopErrorHandler()

    if success != 0 and success is not None:

        opened = ogr.Open(out_path)

        if opened is None:
            raise RuntimeError("Error while clipping geometry.")

        layer = opened.GetLayer()
        features = layer.GetFeatureCount()

        if features == 0:
            warn("Error while clipping geometry. No features in output.", RuntimeWarning)

        return out_path
    else:
        raise RuntimeError("Error while clipping geometry.")


def vector_clip(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    clip_geom: Union[str, ogr.DataSource],
    out_path: Optional[str] = None,
    to_extent: bool = False,
    target_projection: Optional[Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]] = None,
    preserve_fid: bool = True,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    overwrite: bool = True,
    promote_to_multi: bool = True,
    verbose: bool = False,
) -> Union[str, List[str]]:
    """
    Clips a vector to a geometry.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]]
        Vector(s) to clip.

    clip_geom : Union[str, ogr.DataSource]
        Vector to clip with.

    out_path : Optional[str], optional
        Output path. If None, memory vectors are created. Default: None
    
    to_extent : bool, optional
        Clip to extent. Default: False

    target_projection : Optional[Union[str, int, gdal.Dataset, ogr.DataSource, osr.SpatialReference]], optional
        Target projection. Default: None
    
    preserve_fid : bool, optional
        Preserve fid. Default: True

    prefix : str, optional
        Prefix to add to the output path. Default: ""

    suffix : str, optional
        Suffix to add to the output path. Default: ""

    add_uuid : bool, optional
        Add a uuid to the output path. Default: False

    overwrite : bool, optional
        Overwrite output. Default: True

    promote_to_multi : bool, optional
        Promote to multi. Default: True
    
    verbose : bool, optional
        Print progress/warnings. Default: False

    Returns
    -------
    Union[str, List[str]]
        Path to the clipped vector(s)
    """
    utils_base._type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")
    utils_base._type_check(clip_geom, [ogr.DataSource, gdal.Dataset, str, list, tuple], "clip_geom")
    utils_base._type_check(out_path, [str, None], "out_path")
    utils_base._type_check(to_extent, [bool], "to_extent")
    utils_base._type_check(target_projection, [str, ogr.DataSource, gdal.Dataset, osr.SpatialReference, int, None], "target_projection")
    utils_base._type_check(preserve_fid, [bool], "preserve_fid")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(promote_to_multi, [bool], "promote_to_multi")
    utils_base._type_check(verbose, [bool], "verbose")

    input_is_list = isinstance(vector, list)
    input_data = utils_io._get_input_paths(vector, "vector")

    output_data = utils_io._get_output_paths(
        input_data,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        overwrite=overwrite,
    )

    utils_path._delete_if_required_list(output_data, overwrite)

    output = []
    for idx, in_vector in enumerate(input_data):
        output.append(
            _vector_clip(
                in_vector,
                clip_geom,
                out_path=output_data[idx],
                to_extent=to_extent,
                target_projection=target_projection,
                preserve_fid=preserve_fid,
                promote_to_multi=promote_to_multi,
                overwrite=overwrite,
                verbose=verbose,
            )
        )

    if input_is_list:
        return output

    return output[0]



"""
### Rasterize vectors. ###

Functions to rasterize vectors to rasters.

TODO:
    * Add support for projections
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
from osgeo import gdal, ogr, osr
import numpy as np

# Internal
from buteo.utils import (
    utils_gdal,
    utils_path,
    utils_projection,
    utils_translate,
)
from buteo.raster.reproject import raster_reproject
from buteo.vector.reproject import vector_reproject
from buteo.vector.metadata import _vector_to_metadata
from buteo.vector import core_vector
from buteo.raster import core_raster



def vector_rasterize(
    vector: Union[str, ogr.DataSource],
    pixel_size: Union[float, int],
    out_path: Optional[str] = None,
    extent: Optional[List[Union[float, int]]] = None,
    projection: Optional[Union[str, int, osr.SpatialReference, gdal.Dataset, ogr.DataSource]] = None,
    all_touch: bool = False,
    dtype: Union[str, np.dtype] = "uint8",
    optim: str = "raster",
    band: int = 1,
    fill_value: Union[int, float] = 0,
    nodata_value: Optional[Union[int, float]] = None,
    check_memory: bool = True,
    burn_value=1,
    attribute=None,
) -> str:
    """
    Rasterize a vector to a raster.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource]
        The vector to rasterize.
    
    pixel_size : Union[float, int]
        The pixel size of the raster.
    
    out_path : Optional[str], optional
        Path to output raster. Default: None

    extent: Optional[List[Union[float, int, str, gdal.Dataset, ogr.DataSource]]], optional
        Extent of raster. Default: None

    projection : Optional[Union[str, int, osr.SpatialReference, gdal.Dataset, ogr.DataSource]], optional
        Projection of raster. Default: None

    all_touch : bool, optional
        All pixel touch? Default: False

    dtype : Union[str, np.dtype], optional
        Data type of raster. Default: "uint8"

    optim : str, optional
        Optimization for raster or vector? Default: "raster"

    band : int, optional
        Band to rasterize. Default: 1
    
    fill_value : Union[int, float], optional
        Fill value. Default: 0

    nodata_value : Optional[Union[int, float]], optional
        Nodata value. Default: None

    check_memory : bool, optional
        Check memory? Default: True

    burn_value : int, optional
        Burn value (The value to burn into the raster). Default: 1

    attribute : Optional[str], optional
        Attribute to rasterize. Default: None
        This will burn the value of the attribute into the raster.

    Returns
    -------
    str
        Path to output raster.
    """
    vector_fn = vector

    if out_path is None:
        out_path = utils_path._get_temp_filepath(vector, suffix="_rasterized", ext="tif")

    if projection is not None:
        projection = utils_projection.parse_projection(projection)

        vector_fn = vector_reproject(vector, projection)

        if isinstance(extent, (gdal.Dataset, ogr.DataSource)):
            if utils_gdal._check_is_raster(extent):
                extent = raster_reproject(extent, projection, add_uuid=True, suffix="_reprojected")
            else:
                extent = vector_reproject(extent, projection, add_uuid=True, suffix="_reprojected")

        if isinstance(pixel_size, (gdal.Dataset, ogr.DataSource)):
            if utils_gdal._check_is_raster(pixel_size):
                pixel_size = raster_reproject(pixel_size, projection, add_uuid=True, suffix="_reprojected")
            else:
                pixel_size = vector_reproject(pixel_size, projection, add_uuid=True, suffix="_reprojected")

    if projection is None and isinstance(pixel_size, (gdal.Dataset, ogr.DataSource)):
        if extent is not None:
            extent = raster_reproject(extent, pixel_size, add_uuid=True, suffix="_reprojected")

        vector_fn = vector_reproject(vector, pixel_size, add_uuid=True, suffix="_reprojected")

    # Open the data source and read in the extent
    source_ds = core_vector._vector_open(vector_fn)
    source_meta = _vector_to_metadata(vector_fn)
    source_layer = source_ds.GetLayer()
    x_min, x_max, y_min, y_max = source_layer.GetExtent()

    if isinstance(pixel_size, (gdal.Dataset, str)):
        pixel_size_x = core_raster._raster_open(pixel_size).GetGeoTransform()[1]
        pixel_size_y = abs(core_raster._raster_open(pixel_size).GetGeoTransform()[5])
    elif isinstance(pixel_size, (int, float)):
        pixel_size_x = pixel_size
        pixel_size_y = pixel_size
    elif isinstance(pixel_size, (list, tuple)):
        pixel_size_x, pixel_size_y = pixel_size
    else:
        raise ValueError("pixel_size must be either a gdal.Dataset or a tuple of (x, y)")

    # Create the destination data source
    x_res = int((x_max - x_min) / pixel_size_x)
    y_res = int((y_max - y_min) / pixel_size_y)

    if extent is not None:
        extent_vector = _vector_to_metadata(extent)
        extent_dict = extent_vector["bbox_dict"]
        x_res = int((extent_dict["x_max"] - extent_dict["x_min"]) / abs(pixel_size_x))
        y_res = int((extent_dict["y_max"] - extent_dict["y_min"]) / abs(pixel_size_y))
        x_min = extent_dict["x_min"]
        y_max = extent_dict["y_max"]

    if check_memory is False:
        gdal.SetConfigOption("CHECK_DISK_FREE_SPACE", "FALSE")

    if x_res == 0.0 or y_res == 0.0:
        raise ValueError("Invalid extent. x_res or y_res is 0.0. Did you consider latlng coordinates?")

    try:
        target_ds = gdal.GetDriverByName("GTiff").Create(
            out_path,
            x_res,
            y_res,
            1,
            utils_translate._translate_dtype_numpy_to_gdal(dtype),
        )
    finally:
        gdal.SetConfigOption("CHECK_DISK_FREE_SPACE", "TRUE")

    if target_ds is None:
        raise RuntimeError("Unable to rasterize.")

    target_ds.SetGeoTransform((x_min, pixel_size_x, 0, y_max, 0, -1 * abs(pixel_size_y)))
    target_ds.SetProjection(source_meta["projection_wkt"])

    band = target_ds.GetRasterBand(1)

    if nodata_value is not None:
        band.SetNoDataValue(nodata_value)
    else:
        band.Fill(fill_value)

    options = []
    if all_touch is True:
        options.append("ALL_TOUCHED=TRUE")
    else:
        options.append("ALL_TOUCHED=FALSE")

    if optim == "raster":
        options.append("OPTIM=RASTER")
    elif optim == "vector":
        options.append("OPTIM=VECTOR")
    else:
        options.append("OPTIM=AUTO")

    if attribute is None:
        gdal.RasterizeLayer(
            target_ds,
            [1],
            source_layer,
            burn_values=[burn_value],
            options=options,
        )
    else:
        options.append(f"ATTRIBUTE={attribute}")
        gdal.RasterizeLayer(
            target_ds, [1], source_layer, options=utils_gdal._get_default_creation_options(options)
        )

    return out_path



"""
### Reproject rasters. ###

Module to reproject rasters to a target coordinate reference system.
Can uses references from vector or other raster datasets.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
from osgeo import gdal, ogr, osr
import numpy as np

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_gdal,
    utils_path,
    utils_projection,
    utils_translate,
)
from buteo.raster import core_raster


def _find_common_projection(
    rasters: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
):
    """
    Finds the common projection of a list of rasters.

    If no rasters have the majority of the same projection, the function will return the
    projection of the first raster. If only one raster is provided, the projection of that
    raster will be returned.

    Parameters
    ----------
    rasters : Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]]
        A list of rasters.

    Returns
    -------
    osr.SpatialReference
        The common projection.
    """
    utils_base._type_check(rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "rasters")

    rasters = utils_base._get_variable_as_list(rasters)

    assert utils_gdal._check_is_raster_list(rasters), "rasters must be a list of rasters."

    # Get the projection of each raster
    projections = [utils_projection.parse_projection(raster) for raster in rasters]

    # Get the most common projection
    common_projection = max(set(projections), key=projections.count)

    return common_projection


def _raster_reproject(
    raster: Union[str, gdal.Dataset],
    projection: Union[int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference],
    out_path: Optional[Union[str, List[str]]] = None,
    *,
    resample_alg: str = "nearest",
    copy_if_same: bool = True,
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
    dst_nodata: Union[str, int, float] = "infer",
    dtype: Optional[str] = None,
    prefix: str = "",
    suffix: str = "reprojected",
    add_uuid: bool = False,
    add_timestamp: bool = True,
    memory: float = 0.8,
) -> Union[str, List[str]]:
    """ Internal. """
    assert isinstance(raster, (gdal.Dataset, str)), f"The input raster must be in the form of a str or a gdal.Dataset: {raster}"

    if out_path is None:
        out_path = utils_path._get_temp_filepath(
            name="reprojected_raster.tif",
            prefix=prefix,
            suffix=suffix,
            add_uuid=add_uuid,
            add_timestamp=add_timestamp,
        )
    else:
        if not utils_path._check_is_valid_output_filepath(out_path):
            raise ValueError(f"Invalid output path: {out_path}")

    ref = core_raster._raster_open(raster)
    metadata = core_raster._get_basic_metadata_raster(ref)

    out_format = utils_gdal._get_raster_driver_name_from_path(out_path)

    original_projection = utils_projection.parse_projection(ref)
    target_projection = utils_projection.parse_projection(projection)

    if not isinstance(original_projection, osr.SpatialReference):
        raise RuntimeError("Error while parsing input projection.")

    if not isinstance(target_projection, osr.SpatialReference):
        raise RuntimeError("Error while parsing target projection.")

    # The input is already in the correct projection.
    if not copy_if_same and utils_projection._check_projections_match(original_projection, target_projection):
        return utils_gdal._get_path_from_dataset(ref)

    src_nodata = metadata["nodata_value"]
    out_nodata = None
    if dst_nodata == "infer":
        dst_nodata = src_nodata
    else:
        assert isinstance(dst_nodata, (int, float, type(None))), "dst_nodata must be an int, float, 'infer', or 'None'"
        out_nodata = dst_nodata

    if dtype is None:
        dtype = metadata["dtype"]

    utils_path._delete_if_required(out_path, overwrite)

    # Translate for GdalWarp
    resample_alg = utils_translate._translate_resample_method(resample_alg)
    creation_options = utils_gdal._get_default_creation_options(creation_options)
    memory_limit = utils_gdal._get_dynamic_memory_limit(memory)
    output_dtype = utils_translate._translate_dtype_numpy_to_gdal(
        utils_translate._parse_dtype(dtype),
    )
    src_projection = original_projection.ExportToWkt()
    dst_projection = target_projection.ExportToWkt()

    warp_options = gdal.WarpOptions(
        format=out_format,
        srcSRS=src_projection,
        dstSRS=dst_projection,
        resampleAlg=resample_alg,
        outputType=output_dtype,
        creationOptions=creation_options,
        srcNodata=src_nodata,
        dstNodata=out_nodata,
        warpMemoryLimit=memory_limit,
        multithread=True,
    )

    reprojected = gdal.Warp(out_path, ref, options=warp_options)

    if reprojected is None:
        raise RuntimeError(f"Error while reprojecting raster: {raster}")

    return out_path


def raster_reproject(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    projection: Union[int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference],
    out_path: Optional[Union[str, List[str]]] = None,
    *,
    resample_alg: str = "nearest",
    copy_if_same: bool = True,
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
    dst_nodata: Union[str, int, float] = "infer",
    dtype: Optional[str] = None,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
) -> Union[str, List[str]]:
    """
    Reproject raster(s) to a target coordinate reference system.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]]
        The raster(s) to reproject.

    projection : Union[int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference]
        The projection to reproject the raster to. The input can be a WKT proj,
        EPSG proj, Proj, osr proj, or read from a vector or raster datasource
        either from path or in-memory.

    out_path : Optional[Union[str, List[str]]], optional
        The output path, default: None. If not provided, the output path is inferred
        from the input.

    resample_alg : str, optional
        The resampling algorithm, default: "nearest".

    copy_if_same : bool, optional
        If the input and output projections are the same, copy the input raster to the
        output path, default: True.

    overwrite : bool, optional
        If the output path already exists, overwrite it, default: True.

    creation_options : Optional[List[str]], optional
        A list of creation options for the output raster, default: None.

    dst_nodata : Union[str, int, float], optional
        The nodata value for the output raster, default: "infer".

    dtype : Optional[str], optional
        The data type for the output raster, default: None.

    prefix : str, optional
        The prefix to add to the output path, default: "".

    suffix : str, optional
        The suffix to add to the output path, default: "".

    add_uuid : bool, optional
        If True, add a UUID to the output path, default: False.

    Returns
    -------
    Union[str, List[str]]
        The output path(s).
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(projection, [int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference], "projection")
    utils_base._type_check(out_path, [list, str, None], "out_path")
    utils_base._type_check(resample_alg, [str], "resample_alg")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(dst_nodata, [str, int, float], "dst_nodata")
    utils_base._type_check(dtype, [str, None, np.dtype, type(np.int8)], "dtype")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "postfix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")

    input_is_list = isinstance(raster, list)

    raster_list = utils_io._get_input_paths(raster, "raster")
    out_path_list = utils_io._get_output_paths(
        raster_list,
        out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        change_ext="tif",
    )

    utils_path._delete_if_required_list(out_path_list, overwrite)

    output = []
    for index, in_raster in enumerate(raster_list):
        output.append(
            _raster_reproject(
                in_raster,
                projection,
                out_path=out_path_list[index],
                resample_alg=resample_alg,
                copy_if_same=copy_if_same,
                overwrite=overwrite,
                creation_options=creation_options,
                dst_nodata=dst_nodata,
                dtype=dtype,
                prefix=prefix,
                suffix=suffix,
            )
        )

    if input_is_list:
        return output

    return output[0]



"""
### Merge vectors. ###

Merges vectors into a single vector file.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, List, Optional

# External
from osgeo import ogr
from osgeo_utils.ogrmerge import ogrmerge

# Internal
from buteo.utils import (
    utils_base,
    utils_gdal,
    utils_path,
    utils_projection,
)
from buteo.vector import core_vector
from buteo.vector.metadata import _vector_to_metadata



def vector_merge_layers(
    vectors: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    out_path: Optional[str] = None,
) -> str:
    """
    Merge vectors to a single geopackage.

    Parameters
    ----------
    vectors : Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]]
        The input vectors.

    out_path : Optional[str], optional
        The output path, default: None.

    Returns
    -------
    str
        The output path.
    """
    utils_base._type_check(vectors, [[str, ogr.DataSource]], "vector")
    utils_base._type_check(out_path, [str, [str], None], "out_path")

    vector_list = utils_base._get_variable_as_list(vectors)

    assert utils_gdal._check_is_vector_list(vector_list), "Invalid input vector list"

    if out_path is None:
        out_path = utils_path._get_temp_filepath(vector_list[0], suffix="_merged")

    driver = ogr.GetDriverByName(utils_gdal._get_vector_driver_name_from_path(out_path))

    merged_ds = driver.CreateDataSource(out_path)

    for vector in vector_list:
        ref = core_vector._vector_open(vector)
        metadata = _vector_to_metadata(ref)

        for layer in metadata["layers"]:
            name = layer["layer_name"]
            merged_ds.CopyLayer(ref.GetLayer(name), name, ["OVERWRITE=YES"])

    merged_ds.FlushCache()

    return out_path


def vector_merge_features(
    vectors: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    out_path: Optional[str] = None,
    projection: Optional[str] = None,
    single_layer: bool = True,
    overwrite: bool = True,
    skip_failures: bool = True,
) -> str:
    """
    Merge vectors to a single geopackage.

    Parameters
    ----------
    vectors : Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]]
        The input vectors.

    out_path : Optional[str], optional
        The output path, default: None.

    projection : Optional[str], optional
        The projection of the output vector, default: None.
        If None, the projection of the first input vector is used.

    single_layer : bool, optional
        If True, all layers are merged into a single layer, default: True.

    overwrite : bool, optional
        If True, the output file is overwritten if it exists, default: True.
    
    skip_failures : bool, optional
        If True, failures are skipped, default: True.

    Returns
    -------
    str
        The output path.
    """
    utils_base._type_check(vectors, [[str, ogr.DataSource]], "vector")
    utils_base._type_check(out_path, [str, [str], None], "out_path")

    vector_list = utils_base._get_variable_as_list(vectors)

    assert utils_gdal._check_is_vector_list(vector_list), "Invalid input vector list"

    if out_path is None:
        out_path = utils_path._get_temp_filepath(vector_list[0], suffix="_merged")

    driver_name = utils_gdal._get_driver_name_from_path(out_path)

    if projection is None:
        target_projection = utils_projection._get_projection_from_vector(vector_list[0]).ExportToWkt()
    else:
        target_projection = utils_projection.parse_projection(projection, return_wkt=True)

    success = ogrmerge(
        vector_list,
        out_path,
        single_layer=single_layer,
        driver_name=driver_name,
        t_srs=target_projection,
        overwrite_ds=overwrite,
        skip_failures=skip_failures,
        src_layer_field_name="og_name",
        src_layer_field_content="{LAYER_NAME}"
    )

    return out_path



"""
### Vectorize rasters. ###

Module to turn rasters into vector representations.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
from osgeo import gdal, ogr

# Internal
from buteo.utils import (
    utils_gdal,
    utils_path,
    utils_translate,
    utils_projection
)
from buteo.raster import core_raster


def raster_warp(
    raster: Union[str, gdal.Dataset],
    out_path: Optional[str] = None,
    src_projection: Optional[Union[str, gdal.Dataset]] = None,
    dst_projection: Optional[Union[str, gdal.Dataset]] = None,
    resampling_alg: str = "near",
    align_pixels: bool = False,
    dst_extent: Optional[Union[List[Union[int, float]], gdal.Dataset]] = None,
    dst_extent_srs: Optional[Union[str, gdal.Dataset]] = None,
    dst_x_res: Optional[Union[str, gdal.Dataset]] = None,
    dst_y_res: Optional[Union[str, gdal.Dataset]] = None,
    dst_width: Optional[Union[str, gdal.Dataset]] = None,
    dst_height: Optional[Union[str, gdal.Dataset]] = None,
    dst_nodata: Optional[Union[str, gdal.Dataset]] = None,
    clip_geom: Optional[Union[str, ogr.DataSource]]= None,
    overwrite: bool = True,
) -> str:
    """ Warp a raster to a new projection, resolution, extent, or size. """

    if out_path is None:
        out_path = utils_path._get_temp_filepath(raster)

    assert utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite), f"out_path is not a valid output path: {out_path}"

    out_format = utils_gdal._get_driver_name_from_path(out_path)
    input_metadata = core_raster._get_basic_metadata_raster(raster)
    if src_projection is None:
        src_projection = input_metadata["projection_wkt"]
    else:
        src_projection = utils_projection.parse_projection(src_projection, return_wkt=True)

    if dst_projection is None:
        dst_projection = src_projection
    else:
        dst_projection = utils_projection.parse_projection(dst_projection, return_wkt=True)

    if dst_extent_srs is not None:
        dst_extent_srs = utils_projection.parse_projection(dst_extent_srs, return_wkt=True)

    resampling_alg = utils_translate._translate_resample_method(resampling_alg)

    if dst_extent is not None:
        if isinstance(dst_extent, (str, gdal.Dataset)):
            dst_extent = core_raster._get_basic_metadata_raster(dst_extent)["bbox_gdal"]

    if dst_x_res is not None:
        if isinstance(dst_x_res, (str, gdal.Dataset)):
            metadata = core_raster._get_basic_metadata_raster(dst_x_res)
            dst_x_res = metadata["pixel_width"]

    if dst_y_res is not None:
        if isinstance(dst_y_res, (str, gdal.Dataset)):
            metadata = core_raster._get_basic_metadata_raster(dst_y_res)
            dst_y_res = metadata["pixel_height"]

    if dst_width is not None:
        if isinstance(dst_width, (str, gdal.Dataset)):
            metadata = core_raster._get_basic_metadata_raster(dst_width)
            dst_width = metadata["width"]

    if dst_height is not None:
        if isinstance(dst_height, (str, gdal.Dataset)):
            metadata = core_raster._get_basic_metadata_raster(dst_height)
            dst_height = metadata["height"]

    if dst_nodata is not None:
        if isinstance(dst_nodata, (str, gdal.Dataset)):
            metadata = core_raster._get_basic_metadata_raster(dst_nodata)
            dst_nodata = metadata["nodata_value"]

    warp_options = gdal.WarpOptions(
        format=out_format,
        srcSRS=src_projection,
        dstSRS=dst_projection,
        resampleAlg=resampling_alg,
        outputBounds=dst_extent,
        outputBoundsSRS=dst_extent_srs,
        xRes=dst_x_res,
        yRes=dst_y_res,
        width=dst_width,
        height=dst_height,
        dstNodata=dst_nodata,
        multithread=True,
        targetAlignedPixels=align_pixels,
        cutlineDSName=clip_geom,
    )

    success = gdal.Warp(
        out_path,
        raster,
        options=warp_options,
    )

    assert success, f"Failed to warp raster: {raster}"

    return out_path



"""
### Basic IO functions for working with Vectprs ###

The basic module for interacting with vector data
"""

# TODO: More attribute functions
# TODO: Repair vector functions
# TODO: Sanity checks
# TODO: Joins (by attribute, location, summary, etc..)
# TODO: Union, Erase, ..
# TODO: Sampling functions
# TODO: Vector intersects, etc..

# Standard library
import sys; sys.path.append("../../")
import os
from typing import Union, Optional, List, Dict, Any, Callable, Tuple

# External
from osgeo import ogr, gdal, osr

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_gdal,
    utils_bbox,
    utils_path,
    utils_projection,
)


def _vector_open(
    vector: Union[str, ogr.DataSource, gdal.Dataset],
    writeable: bool = True,
    allow_raster: bool = True,
) -> ogr.DataSource:
    """ Internal. """
    assert isinstance(vector, (str, ogr.DataSource)), "vector must be a path or a DataSource"

    opened = None
    if utils_gdal._check_is_vector(vector):
        if isinstance(vector, str):
            gdal.PushErrorHandler("CPLQuietErrorHandler")
            opened = ogr.Open(vector, gdal.GF_Write) if writeable else ogr.Open(vector, gdal.GF_Read)
            gdal.PopErrorHandler()
        elif isinstance(vector, ogr.DataSource):
            opened = vector
        else:
            raise RuntimeError(f"Could not read input vector: {vector}")

    elif allow_raster and utils_gdal._check_is_raster(vector):
        if isinstance(vector, str):
            gdal.PushErrorHandler("CPLQuietErrorHandler")
            opened = gdal.Open(vector, gdal.GF_Write) if writeable else gdal.Open(vector, gdal.GF_Read)
            gdal.PopErrorHandler()
        elif isinstance(vector, gdal.Dataset):
            opened = vector
        else:
            raise RuntimeError(f"Could not read input vector: {vector}")

        bbox = utils_bbox._get_bbox_from_geotransform(
            opened.GetGeoTransform(),
            opened.RasterXSize,
            opened.RasterYSize,
        )

        projection_wkt = opened.GetProjection()
        projection_osr = osr.SpatialReference()
        projection_osr.ImportFromWkt(projection_wkt)

        vector_bbox = utils_bbox._get_vector_from_bbox(bbox, projection_osr)
        opened = ogr.Open(vector_bbox, gdal.GF_Write) if writeable else ogr.Open(vector_bbox, gdal.GF_Read)

    else:
        raise RuntimeError(f"Could not read input vector: {vector}")

    if opened is None:
        raise RuntimeError(f"Could not read input vector: {vector}")

    return opened


def vector_open(
    vector: Union[str, ogr.DataSource, gdal.Dataset, List[Union[str, ogr.DataSource, gdal.Dataset]]],
    writeable: bool = True,
    allow_raster: bool = True,
) -> Union[ogr.DataSource, List[ogr.DataSource]]:
    """
    Opens a vector to an ogr.Datasource class.

    Parameters
    ----------
    vector : str, ogr.DataSource, gdal.Dataset, list[str, ogr.DataSource, gdal.Dataset]
        The vector to open. If a raster is supplied the bounding box is opened as a vector.

    writeable : bool, optional
        If True, the vector is opened in write mode. Default: True

    allow_raster : bool, optional
        If True, a raster will be opened as a vector bounding box. Default: True
    
    allow_lists : bool, optional
        If True, the input can be a list of vectors. Default: True

    Returns
    -------
    ogr.DataSource, list[ogr.DataSource]
        The opened vector(s).
    """
    utils_base._type_check(vector, [str, ogr.DataSource, gdal.Dataset, [str, ogr.DataSource, gdal.Dataset]], "vector")
    utils_base._type_check(writeable, [bool], "writeable")
    utils_base._type_check(allow_raster, [bool], "allow_raster")

    vectors = utils_base._get_variable_as_list(vector)

    output = []
    for element in vectors:
        output.append(_vector_open(element, writeable=writeable, allow_raster=allow_raster))

    if isinstance(vector, list):
        return output

    return output[0]



def _get_basic_metadata_vector(
    vector: Union[str, ogr.DataSource],
) -> Dict[str, Any]:
    """
    Get basic metadata from a vector.

    Parameters
    ----------
    raster : str or ogr.DataSource
        The raster to get the metadata from.

    Returns
    -------
    Dict[str]
        A dictionary with the metadata.
    """
    utils_base._type_check(vector, [str, ogr.DataSource], "vector")

    datasource = _vector_open(vector)

    # Paths
    path = utils_path._get_unix_path(datasource.GetDescription())
    in_memory = utils_path._check_is_valid_mem_filepath(path)

    layers = []
    vector_bbox = None
    layer_count = datasource.GetLayerCount()
    total_features = 0
    for layer_index in range(layer_count):
        layer = datasource.GetLayerByIndex(layer_index)

        x_min, x_max, y_min, y_max = layer.GetExtent()
        layer_bbox = [x_min, x_max, y_min, y_max]
        vector_bbox = layer_bbox

        layer_name = layer.GetName()

        column_fid = layer.GetFIDColumn()
        column_geom = layer.GetGeometryColumn()

        if column_geom == "":
            column_geom = "geom"

        feature_count = layer.GetFeatureCount()

        projection_osr = layer.GetSpatialRef()
        projection_wkt = layer.GetSpatialRef().ExportToWkt()

        if layer_index > 0:
            v_x_min, v_x_max, v_y_min, v_y_max = vector_bbox
            if x_min < v_x_min:
                vector_bbox[0] = x_min
            if x_max > v_x_max:
                vector_bbox[1] = x_max
            if y_min < v_y_min:
                vector_bbox[2] = y_min
            if y_max > v_y_max:
                vector_bbox[3] = y_max

        layer_defn = layer.GetLayerDefn()

        geom_type_ogr = layer_defn.GetGeomType()
        geom_type = ogr.GeometryTypeToName(layer_defn.GetGeomType())

        field_count = layer_defn.GetFieldCount()
        field_names = []
        field_types = []
        field_types_ogr = []

        for field_index in range(field_count):
            field_defn = layer_defn.GetFieldDefn(field_index)
            field_names.append(field_defn.GetName())
            field_type = field_defn.GetType()
            field_types_ogr.append(field_type)
            field_types.append(field_defn.GetFieldTypeName(field_type))

        layer_dict = {
            "layer_name": layer_name,
            "x_min": x_min,
            "x_max": x_max,
            "y_min": y_min,
            "y_max": y_max,
            "column_fid": column_fid,
            "column_geom": column_geom,
            "feature_count": feature_count,
            "projection_wkt": projection_wkt,
            "projection_osr": projection_osr,
            "geom_type": geom_type,
            "geom_type_ogr": geom_type_ogr,
            "field_count": field_count,
            "field_names": field_names,
            "field_types": field_types,
            "field_types_ogr": field_types_ogr,
            "bbox": layer_bbox,
        }

        layers.append(layer_dict)
        total_features += feature_count

    x_min, x_max, y_min, y_max = vector_bbox
    area = (x_max - x_min) * (y_max - y_min)

    bbox_latlng = utils_projection.reproject_bbox(vector_bbox, projection_osr, utils_projection._get_default_projection_osr())
    bounds_latlng = utils_bbox._get_bounds_from_bbox(vector_bbox, projection_osr, wkt=False)
    bounds_area = bounds_latlng.GetArea()
    x_min, x_max, y_min, y_max = vector_bbox
    area = (x_max - x_min) * (y_max - y_min)
    bounds_wkt = bounds_latlng.ExportToWkt()
    bounds_vector_raw = utils_bbox._get_geom_from_bbox(vector_bbox)
    bounds_vector = bounds_vector_raw.ExportToWkt()

    _centroid = bounds_vector_raw.Centroid()
    centroid = (_centroid.GetX(), _centroid.GetY())
    if projection_osr.IsGeographic():
        centroid_latlng = (centroid[1], centroid[0])
    else:
        _centroid_latlng = bounds_latlng.Centroid()
        centroid_latlng = (_centroid_latlng.GetY(), _centroid_latlng.GetX())


    metadata = {
        "path": path,
        "basename": os.path.basename(path),
        "name": os.path.splitext(os.path.basename(path))[0],
        "folder": os.path.dirname(path),
        "ext": os.path.splitext(path)[1],
        "in_memory": in_memory,
        "driver": datasource.GetDriver().GetName(),
        "projection_osr": projection_osr,
        "projection_wkt": projection_wkt,
        "feature_count": total_features,
        "bbox": vector_bbox,
        "bbox_gdal": utils_bbox._get_gdal_bbox_from_ogr_bbox(vector_bbox),
        "bbox_latlng": bbox_latlng,
        "bounds_latlng": bounds_wkt,
        "bounds_vector": bounds_vector,
        "centroid": centroid,
        "centroid_latlng": centroid_latlng,
        "x_min": vector_bbox[0],
        "x_max": vector_bbox[1],
        "y_min": vector_bbox[2],
        "y_max": vector_bbox[3],
        "area_bounds": bounds_area,
        "area": area,
        "layer_count": layer_count,
        "layers": layers,
    }

    layer = None
    datasource = None
    return metadata


def _vector_filter(
    vector: Union[ogr.DataSource, str],
    filter_function: Callable,
    out_path: Optional[str] = None,
    process_layer: int = -1,
    prefix: str = "",
    suffix: str = "",
    overwrite: bool = True,
) -> str:
    """ Internal. """
    assert isinstance(vector, (str, ogr.DataSource)), "vector must be a string or an ogr.DataSource object."
    assert isinstance(filter_function, (type(lambda: True))), "filter_function must be a function."

    metadata = _get_basic_metadata_vector(vector)

    if out_path is None:
        out_path = utils_path._get_temp_filepath(vector, prefix=prefix, suffix=suffix)

    assert utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite), f"out_path is not a valid output path. {out_path}"

    projection = metadata["projection_osr"]

    driver_name = utils_gdal._get_vector_driver_name_from_path(out_path)
    driver = ogr.GetDriverByName(driver_name)

    datasource_destination = driver.CreateDataSource(out_path)
    datasource_original = _vector_open(vector)

    for i in range(metadata["layer_count"]):
        if process_layer != -1 and i != process_layer:
            continue

        features = metadata["layers"][i]["feature_count"]
        field_names = metadata["layers"][i]["field_names"]

        geom_type = metadata["layers"][i]["geom_type_ogr"]

        layer_original = datasource_original.GetLayer(i)
        layer_destination = datasource_destination.CreateLayer(layer_original.GetName(), projection, geom_type)

        for feature in range(features):
            feature = layer_original.GetNextFeature()

            field_values = []
            for field in field_names:
                field_values.append(feature.GetField(field))

            field_dict = {}
            for j, value in enumerate(field_values):
                field_dict[field_names[j]] = value

            if filter_function(field_dict):
                layer_destination.CreateFeature(feature.Clone())

        layer_destination.SyncToDisk()
        layer_destination.ResetReading()
        layer_destination = None

    return out_path


def vector_filter(
    vector: Union[ogr.DataSource, str, List[Union[ogr.DataSource, str]]],
    filter_function: Callable,
    out_path: Optional[Union[str, List[str]]] = None,
    process_layer: int = -1,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    overwrite: bool = True,
) -> Union[str, List[str]]:
    """
    Filters a vector using its attribute table and a function.

    Parameters
    ----------
    vector : Union[ogr.DataSource, str, List[Union[ogr.DataSource, str]]]
        A vector layer(s) or path(s) to a vector file.

    filter_function : Callable
        A function that takes a dictionary of attributes and returns a boolean.
    
    out_path : Optional[str], optional
        Path to the output vector file. If None, a memory vector will be created. default: None

    process_layer : int, optional
        The index of the layer to process. If -1, all layers will be processed. default: -1

    prefix : str, optional
        A prefix to add to the output vector file. default: ""

    suffix : str, optional
        A suffix to add to the output vector file. default: ""

    add_uuid : bool, optional
        If True, a uuid will be added to the output path. default: False

    overwrite : bool, optional
        If True, the output file will be overwritten if it already exists. default: True

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the output vector file(s).
    """
    utils_base._type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")
    utils_base._type_check(filter_function, [type(lambda: True)], "filter_function")
    utils_base._type_check(process_layer, [int], "process_layer")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(overwrite, [bool], "overwrite")

    input_is_list = isinstance(vector, list)
    input_list = utils_io._get_input_paths(vector, "vector")
    output_list = utils_io._get_output_paths(
        input_list,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        overwrite=overwrite,
        change_ext="gpkg",
    )

    utils_path._delete_if_required_list(output_list, overwrite)

    output = []
    for idx, in_vector in enumerate(input_list):
        output.append(_vector_filter(
            in_vector,
            filter_function,
            out_path=output_list[idx],
            process_layer=process_layer,
            prefix=prefix,
            suffix=suffix,
            overwrite=overwrite,
        ))

    if input_is_list:
        return output

    return output[0]


def vector_add_index(
    vector: Union[ogr.DataSource, str, List[Union[ogr.DataSource, str]]],
) -> Union[str, List[str]]:
    """
    Adds a spatial index to the vector in place, if it doesn't have one.

    Parameters
    ----------
    vector : Union[ogr.DataSource, str, List[Union[ogr.DataSource, str]]]
        A vector layer(s) or path(s) to a vector file.

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the input vector file(s).
    """
    utils_base._type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")

    input_is_list = isinstance(vector, list)
    input_list = utils_io._get_input_paths(vector, "vector")

    try:
        for in_vector in input_list:
            metadata = _get_basic_metadata_vector(in_vector)
            ref = _vector_open(in_vector)

            for layer in metadata["layers"]:
                name = layer["layer_name"]
                geom = layer["column_geom"]

                sql = f"SELECT CreateSpatialIndex('{name}', '{geom}') WHERE NOT EXISTS (SELECT HasSpatialIndex('{name}', '{geom}'));"
                ref.ExecuteSQL(sql, dialect="SQLITE")
    except:
        raise RuntimeError(f"Error while creating indices for {vector}") from None

    if input_is_list:
        return input_list

    return input_list[0]


def _vector_get_attribute_table(
    vector: Union[str, ogr.DataSource],
    process_layer: int = 0,
    include_fids: bool = False,
    include_geometry: bool = False,
    include_attributes: bool = True,
    return_header: bool = True,
) -> Union[Dict[str, Any], Tuple[Dict[str, Any]]]:
    """
    Get the attribute table(s) of a vector.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[str, ogr.DataSource]]
        Vector layer(s) or path(s) to vector layer(s).

    process_layer : int, optional
        The layer to process. Default: 0 (first layer).

    include_fids : bool, optional
        If True, will include the FID column. Default: False.

    include_geometry : bool, optional
        If True, will include the geometry column. Default: False.

    include_attributes : bool, optional
        If True, will include the attribute columns. Default: True.

    return_header : bool, optional
        If True, will return the header. Default: True.

    Returns
    -------
    attribute_table : Dict[str, Any]
        The attribute table(s) of the vector(s).
    """
    assert isinstance(vector, (str, ogr.DataSource)), "vector must be a string or an ogr.DataSource object."
    assert isinstance(process_layer, int), "process_layer must be an integer."
    assert isinstance(include_fids, bool), "include_fids must be a boolean."
    assert isinstance(include_geometry, bool), "include_geometry must be a boolean."
    assert isinstance(include_attributes, bool), "include_attributes must be a boolean."
    assert isinstance(return_header, bool), "return_header must be a boolean."

    ref = _vector_open(vector)
    metadata = _get_basic_metadata_vector(ref)

    attribute_table_header = metadata["layers"][process_layer]["field_names"]
    attribute_table = []

    layer = ref.GetLayer(process_layer)
    layer.ResetReading()
    while True:
        feature = layer.GetNextFeature()

        if feature is None:
            break

        attributes = [feature.GetFID()]

        for field_name in attribute_table_header:
            attributes.append(feature.GetField(field_name))

        if include_geometry:
            geom_defn = feature.GetGeometryRef()
            attributes.append(geom_defn.ExportToIsoWkt())

        attribute_table.append(attributes)

    attribute_table_header.insert(0, "fid")

    if include_geometry:
        attribute_table_header.append("geom")

    ref = None
    layer = None

    if return_header:
        return attribute_table_header, attribute_table

    return attribute_table


def vector_get_attribute_table(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    process_layer: int = 0,
    include_fids: bool = False,
    include_geometry: bool = False,
    include_attributes: bool = True,
    return_header: bool = True,
) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    """
    Get the attribute table(s) of a vector.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[str, ogr.DataSource]]
        Vector layer(s) or path(s) to vector layer(s).

    process_layer : int, optional
        The layer to process. Default: 0 (first layer).

    include_fids : bool, optional
        If True, will include the FID column. Default: False.

    include_geometry : bool, optional
        If True, will include the geometry column. Default: False.

    include_attributes : bool, optional
        If True, will include the attribute columns. Default: True.

    return_header : bool, optional
        If True, will return the header. Default: True.
        
    Returns
    -------
    attribute_table : Dict[str, Any]
        The attribute table(s) of the vector(s).
    """
    utils_base._type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")
    utils_base._type_check(process_layer, [int], "process_layer")
    utils_base._type_check(include_fids, [bool], "include_fids")
    utils_base._type_check(include_geometry, [bool], "include_geometry")
    utils_base._type_check(include_attributes, [bool], "include_attributes")
    utils_base._type_check(return_header, [bool], "return_header")

    input_is_list = isinstance(vector, list)
    input_list = utils_io._get_input_paths(vector, "vector")

    output_attributes = []
    output_headers = []
    for in_vector in input_list:
        if return_header:
            header, table = _vector_get_attribute_table(
                in_vector,
                process_layer=process_layer,
                include_fids=include_fids,
                include_geometry=include_geometry,
                include_attributes=include_attributes,
                return_header=True,
            )
            output_headers.append(header)
            output_attributes.append(table)

        else:
            output_attributes.append(_vector_get_attribute_table(
                in_vector,
                process_layer=process_layer,
                include_fids=include_fids,
                include_geometry=include_geometry,
                include_attributes=include_attributes,
                return_header=False,
            ))

    if return_header:
        if input_is_list:
            return output_headers, output_attributes

        return output_headers[0], output_attributes[0]
    
    if input_is_list:
        return output_attributes

    return output_attributes[0]


def vector_filter_layer(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    layer_name_or_idx: Union[str, int],
    out_path: Optional[Union[str, List[str]]] = None,
    prefix: str = "",
    suffix: str = "_layer",
    add_uuid: bool = False,
    overwrite: bool = True,
):
    """
    Filters a multi-layer vector source to a single layer.
    
    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[str, ogr.DataSource]]
        Vector layer(s) or path(s) to vector layer(s).
    
    layer_name_or_idx : Union[str, int]
        The name or index of the layer to filter.

    out_path : Optional[str], optional
        The path to the output vector. If None, will create a new file in the same directory as the input vector. Default: None.

    prefix : str, optional
        Prefix to add to the output vector. Default: "".

    suffix : str, optional
        Suffix to add to the output vector. Default: "_layer".
    
    add_uuid : bool, optional
        If True, will add a UUID to the output vector. Default: False.

    overwrite : bool, optional
        If True, will overwrite the output vector if it already exists. Default: True.

    Returns
    -------
    out_path : str
        Path to the output vector.
    """
    input_is_list = isinstance(vector, list)

    input_list = utils_io._get_input_paths(vector, "vector")
    output_list = utils_io._get_output_paths(
        input_list,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        overwrite=overwrite,
        change_ext="gpkg",
    )

    utils_path._delete_if_required_list(output_list, overwrite)

    output = []
    for idx, in_vector in enumerate(input_list):
        ref = _vector_open(in_vector)
        out_path = output_list[idx]

        if isinstance(layer_name_or_idx, int):
            layer = ref.GetLayerByIndex(layer_name_or_idx)
        elif isinstance(layer_name_or_idx, str):
            layer = ref.GetLayer(layer_name_or_idx)
        else:
            raise RuntimeError("Wrong datatype for layer selection")

        driver_name = utils_gdal._get_vector_driver_name_from_path(out_path)
        driver = ogr.GetDriverByName(driver_name)

        destination = driver.CreateDataSource(out_path)
        destination.CopyLayer(layer, layer.GetName(), ["OVERWRITE=YES"])
        destination.FlushCache()

        destination = None
        ref = None

        output.append(out_path)

    if input_is_list:
        return output

    return output[0]


def vector_to_extent(
    vector: Union[str, ogr.DataSource],
    out_path: Optional[str] = None,
    *,
    latlng: bool = False,
    overwrite: bool = True,
) -> str:
    """
    Converts a vector to a vector file with the extent as a polygon.

    Parameters
    ----------
    vector : str or ogr.DataSource
        The vector to convert.

    out_path : str, optional
        The path to save the extent to. If None, the extent is saved in memory. Default: None.

    latlng : bool, optional
        If True, the extent is returned in latlng coordinates. If false,
        the projection of the vector is used. Default: False.

    overwrite : bool, optional
        If True, the output file is overwritten if it exists. Default: True.

    Returns
    -------
    str
        The path to the extent.
    """
    utils_base._type_check(vector, [str, ogr.DataSource], "vector")
    utils_base._type_check(out_path, [str, type(None)], "out_path")
    utils_base._type_check(latlng, [bool], "latlng")
    utils_base._type_check(overwrite, [bool], "overwrite")

    if out_path is None:
        out_path = utils_path._get_temp_filepath("temp_extent.gpkg", add_timestamp=True, add_uuid=True)

    if not utils_path._check_is_valid_output_filepath(out_path):
        raise ValueError(f"Invalid output path: {out_path}")

    metadata = _get_basic_metadata_vector(vector)

    if latlng:
        extent = metadata["bounds_latlng"]
    else:
        extent = metadata["bounds_vector"]

    extent = ogr.CreateGeometryFromWkt(extent, metadata["projection_osr"])

    driver_name = utils_gdal._get_driver_name_from_path(out_path)

    driver = ogr.GetDriverByName(driver_name)
    extent_ds = driver.CreateDataSource(out_path)
    extent_layer = extent_ds.CreateLayer("extent", metadata["projection_osr"], ogr.wkbPolygon)
    extent_feature = ogr.Feature(extent_layer.GetLayerDefn())
    extent_feature.SetGeometry(extent)
    extent_layer.CreateFeature(extent_feature)

    extent_ds = None
    extent_layer = None
    extent_feature = None

    return out_path


def vector_copy(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    out_path: Optional[Union[str, List[str]]] = None,
) -> Union[str, List[str]]:
    """
    Creates a copy of a vector.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[str, ogr.DataSource]]
        Vector layer(s) or path(s) to vector layer(s).

    out_path : Optional[str], optional
        The path to the output vector. If None, will create a new file in the same directory as the input vector. Default: None.

    Returns
    -------
    out_path : str
        Path to the output vector.
    """
    assert isinstance(vector, (str, ogr.DataSource)), "vector must be a string or an ogr.DataSource object."
    assert isinstance(out_path, (str, list, type(None))), "out_path must be a string, list, or None."

    input_is_list = isinstance(vector, list)

    input_list = utils_io._get_input_paths(vector, "vector")
    output_list = utils_io._get_output_paths(input_list, out_path)

    utils_path._delete_if_required_list(output_list, overwrite=True)

    output = []
    for idx, in_vector in enumerate(input_list):
        ref = _vector_open(in_vector)

        driver_name = utils_gdal._get_vector_driver_name_from_path(output_list[idx])
        driver = ogr.GetDriverByName(driver_name)

        destination = driver.CreateDataSource(output_list[idx])
        layers = ref.GetLayerCount()

        for layer_index in range(layers):
            layer = ref.GetLayer(layer_index)
            layer.ResetReading()
            destination.CopyLayer(layer, layer.GetName(), ["OVERWRITE=YES"])

        destination.FlushCache()
        destination = None
        ref = None

        output.append(output_list[idx])

    if input_is_list:
        return output

    return output[0]


def vector_reset_fids(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
) -> Union[str, List[str]]:
    """ 
    Resets the FID column of a vector to 0, 1, 2, 3, ...

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[str, ogr.DataSource]]
        Vector layer(s) or path(s) to vector layer(s).

    out_path : Optional[str], optional
        The path to the output vector. If None, will create a new file in the same directory as the input vector. Default: None.

    Returns
    -------
    out_path : str
        Path to the output vector.
    """
    assert isinstance(vector, (str, ogr.DataSource)), "vector must be a string or an ogr.DataSource object."

    input_is_list = isinstance(vector, list)

    input_list = utils_io._get_input_paths(vector, "vector")

    output = []
    for idx, in_vector in enumerate(input_list):
        ref = _vector_open(in_vector)

        layers = ref.GetLayerCount()

        for layer_index in range(layers):
            layer = ref.GetLayer(layer_index)
            layer.ResetReading()
            fids = []

            for feature in layer:
                fids.append(feature.GetFID())

            layer.ResetReading()
            fids = sorted(fids)

            layer_defn = layer.GetLayerDefn()
            field_names = [layer_defn.GetFieldDefn(i).GetName() for i in range(layer_defn.GetFieldCount())]

            for feature in layer:
                current_fid = feature.GetFID()
                target_fid = fids.index(current_fid)

                # If there is a fid field, update it too.
                if "fid" in field_names:
                    feature.SetField("fid", str(target_fid))

                feature.SetFID(target_fid)
                layer.SetFeature(feature)

            layer.SyncToDisk()

        ref = None

        output.append(input_list[idx])

    if input_is_list:
        return output

    return output[0]


def vector_from_bbox(
    bbox: List[Union[float, int]],
    projection: Union[str, osr.SpatialReference, None] = None,
    out_path: Optional[str] = None,
) -> str:
    """
    Creates a vector from a bounding box.
    
    Parameters
    ----------
    bbox : List[float, int]
        The bounding box to create the vector from. Must be in the format [x_min, x_max, y_min, y_max].

    projection : Union[str, osr.SpatialReference, None], optional
        The projection of the bounding box. If None, the default (latlng) projection is used. Default: None.

    out_path : Optional[str], optional
        The path to the output vector. If None, will create a new file in the same directory as the input vector. Default: None.
    """
    assert utils_bbox._check_is_valid_bbox(bbox), "bbox is not a valid bounding box."

    if projection is None:
        proj = utils_projection._get_default_projection_osr()
    else:
        proj = utils_projection.parse_projection(projection)

    if utils_projection._projection_is_latlng(proj):
        bbox = [bbox[2], bbox[3], bbox[0], bbox[1]]

    if out_path is None:
        out_path = utils_path._get_temp_filepath("temp_bbox.gpkg", add_timestamp=True, add_uuid=True)
    
    if not utils_path._check_is_valid_output_filepath(out_path):
        raise ValueError(f"Invalid output path: {out_path}")
    
    driver_name = utils_gdal._get_vector_driver_name_from_path(out_path)
    driver = ogr.GetDriverByName(driver_name)

    datasource = driver.CreateDataSource(out_path)
    layer = datasource.CreateLayer("bbox", geom_type=ogr.wkbPolygon, srs=proj)
    feature = ogr.Feature(layer.GetLayerDefn())
    feature.SetGeometry(utils_bbox._get_geom_from_bbox(bbox))
    
    layer.CreateFeature(feature)
    layer.SyncToDisk()

    datasource = None
    layer = None
    feature = None

    return out_path


def vector_from_points(
    points: List[List[Union[float, int]]],
    projection: Union[str, osr.SpatialReference, None] = None,
    out_path: Optional[str] = None,
    reverse_xy_order: bool = False,
) -> str:
    """
    Creates a point vector from a list of points.
    
    Parameters
    ----------
    points : List[List[float, int]]
        The points to create the vector from. Must be in the format [[x1, y1], [x2, y2], ...].

    projection : Union[str, osr.SpatialReference, None], optional
        The projection of the points. If None, the default (latlng) projection is used. Default: None.

    out_path : Optional[str], optional
        The path to the output vector. If None, will create a new file in the same directory as the input vector. Default: None.
    """
    assert isinstance(points, list), "points must be a list."
    assert len(points) > 0, "points must have at least one point."
    for point in points:
        assert isinstance(point, list), "points must be a list of lists."
        assert len(point) == 2, "points must be a list of lists of length 2."

    if projection is None:
        proj = utils_projection._get_default_projection_osr()
    else:
        proj = utils_projection.parse_projection(projection)

    if out_path is None:
        out_path = utils_path._get_temp_filepath("temp_points.gpkg", add_timestamp=True, add_uuid=True)
    
    if not utils_path._check_is_valid_output_filepath(out_path):
        raise ValueError(f"Invalid output path: {out_path}")
    
    driver_name = utils_gdal._get_vector_driver_name_from_path(out_path)
    driver = ogr.GetDriverByName(driver_name)

    datasource = driver.CreateDataSource(out_path)
    layer = datasource.CreateLayer("points", geom_type=ogr.wkbPoint, srs=proj)
    layer_defn = layer.GetLayerDefn()

    for point in points:
        feature = ogr.Feature(layer_defn)
        geom = ogr.Geometry(ogr.wkbPoint)
        if reverse_xy_order:
            geom.AddPoint(point[1], point[0])
        else:
            geom.AddPoint(point[0], point[1])
        feature.SetGeometry(geom)
        
        layer.CreateFeature(feature)
        feature = None

    layer.GetExtent()
    layer.SyncToDisk()

    datasource = None
    layer = None

    return out_path


def vector_add_field(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    field_name: str,
    field_type: str,
) -> Union[str, List[str]]:
    """
    Adds a field to a vector in place.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]]
        Vector layer(s) or path(s) to vector layer(s).

    field : str
        The name of the field to add.

    field_type : str
        The type of the field to add.
        `['int', 'integer', 'float', 'double', 'string', 'date', 'datetime', 'time', 'binary', 'intlist', 'integerlist', 'floatlist', 'doublelist', 'stringlist', 'datelist', 'datetimelist', 'timelist', 'binarylist']`

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the output vector file(s).
    """
    assert isinstance(vector, (str, ogr.DataSource)), "vector must be a string or an ogr.DataSource object."
    assert isinstance(field_name, str), "field must be a string."
    assert isinstance(field_type, str), "field_type must be a valid vector type."

    if field_type in ['int', 'integer']:
        field_type = ogr.OFTInteger
    elif field_type in ['float', 'double']:
        field_type = ogr.OFTReal
    elif field_type in ['string']:
        field_type = ogr.OFTString
    elif field_type in ['date', 'datetime', 'time']:
        field_type = ogr.OFTDateTime
    elif field_type in ['binary']:
        field_type = ogr.OFTBinary
    elif field_type in ['intlist', 'integerlist']:
        field_type = ogr.OFTIntegerList
    elif field_type in ['floatlist', 'doublelist']:
        field_type = ogr.OFTRealList
    elif field_type in ['stringlist']:
        field_type = ogr.OFTStringList
    else:
        raise ValueError(f"Invalid field_type: {field_type}")

    input_is_list = isinstance(vector, list)

    input_list = utils_io._get_input_paths(vector, "vector")

    for idx, in_vector in enumerate(input_list):
        ref = _vector_open(in_vector)

        layers = ref.GetLayerCount()

        for layer_index in range(layers):
            layer = ref.GetLayer(layer_index)
            layer.ResetReading()

            layer.CreateField(ogr.FieldDefn(field_name, field_type))

            layer.SyncToDisk()

        ref = None

    if input_is_list:
        return input_list
    
    return input_list[0]


def vector_set_attribute_table(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    header: List[str],
    attribute_table: List[List[Any]],
    match: Optional[str] = 'fid',
) -> Union[str, List[str]]:
    """
    Sets the attribute table of a vector in place.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]]
        Vector layer(s) or path(s) to vector layer(s).

    header : List[str]
        The header of the attributes to update in the table.

    attribute_table : List[List[Any]]
        The attributes to update in the table.

    match_fids : bool, optional
        If True, will match the FIDs of the input vector and the attribute table. Default: False.

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the output vector file(s).
    """
    assert isinstance(vector, (str, ogr.DataSource)), "vector must be a string or an ogr.DataSource object."
    assert isinstance(header, list), "header must be a list."
    assert isinstance(attribute_table, list), "attribute_table must be a list."
    assert len(header) == len(attribute_table[0]), "header and attribute_table must have the same number of columns."
    assert isinstance(match, (str, type(None))), "match must be a string or None."

    if match is not None:
        assert match in header, "match must be in header."

    match_idx = header.index(match)

    input_is_list = isinstance(vector, list)
    input_list = utils_io._get_input_paths(vector, "vector")

    for in_vector in input_list:
        ref = _vector_open(in_vector)

        layers = ref.GetLayerCount()

        for layer_index in range(layers):
            layer = ref.GetLayer(layer_index)
            layer.ResetReading()

            layer_defn = layer.GetLayerDefn()
            field_names = [layer_defn.GetFieldDefn(i).GetName() for i in range(layer_defn.GetFieldCount())]

            for field in header:
                if field not in field_names and field != 'fid':
                    layer.CreateField(ogr.FieldDefn(field, ogr.OFTString))

            if match is not None:
                for attr in attribute_table:
                    attr_id = attr[match_idx]
                    feature = layer.GetFeature(attr_id)

                    for field_idx, field in enumerate(header):
                        if field == match:
                            continue

                        feature.SetField(field, attr[field_idx])

                    layer.SetFeature(feature)

            else:
                feature_idx = 0
                for feature in layer:
                    for field_idx, field in enumerate(header):
                        feature.SetField(field, attribute_table[feature_idx][field_idx])
                        feature_idx += 1

                    layer.SetFeature(feature)

            layer.SyncToDisk()

        ref = None

    if input_is_list:
        return input_list
    
    return input_list[0]


def vector_delete_fields(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    fields: List[str],
) -> Union[str, List[str]]:
    """
    Deletes fields from a vector in place.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]]
        Vector layer(s) or path(s) to vector layer(s).

    fields : List[str]
        The fields to delete.

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the output vector file(s).
    """
    assert isinstance(vector, (str, ogr.DataSource)), "vector must be a string or an ogr.DataSource object."
    assert isinstance(fields, list), "fields must be a list."

    input_is_list = isinstance(vector, list)
    input_list = utils_io._get_input_paths(vector, "vector")

    for in_vector in input_list:
        ref = _vector_open(in_vector)

        layers = ref.GetLayerCount()

        for layer_index in range(layers):
            layer = ref.GetLayer(layer_index)
            layer.ResetReading()

            layer_defn = layer.GetLayerDefn()
            field_names = [layer_defn.GetFieldDefn(i).GetName() for i in range(layer_defn.GetFieldCount())]

            for field in fields:
                if field in field_names:
                    layer.DeleteField(field)

            layer.SyncToDisk()

        ref = None

    if input_is_list:
        return input_list
    
    return input_list[0]


"""
### Functions for changing the datatype of a raster. ###
"""

# Standard library
import sys; sys.path.append("../../")
import os
from typing import Union, List, Optional
from warnings import warn

# External
import numpy as np
from osgeo import gdal

# Internal
from buteo.utils import (
    utils_io,
    utils_gdal,
    utils_base,
    utils_path,
    utils_translate,
)
from buteo.raster.core_raster import _raster_open, _get_basic_metadata_raster


def _raster_set_datatype(
    raster: Union[str, gdal.Dataset],
    dtype_str: str,
    out_path: Optional[str] = None,
    *,
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
) -> str:
    """
    Internal function.

    For internal functions, only basic testing should be done.
    The input should be checked in the public function calling the internal function.
    an input and output raster should be provided.
    """
    assert isinstance(raster, (str, gdal.Dataset)), "raster must be a string or a GDAL.Dataset."
    assert isinstance(dtype_str, (str, int, np.dtype)), "dtype_str must be a string."
    assert isinstance(out_path, (str, None)), "out_path must be a string or None."

    if out_path is None:
        out_path = utils_path._get_temp_filepath("set_datatype.tif")

    ref = _raster_open(raster)
    metadata = _get_basic_metadata_raster(ref)

    driver_name = utils_gdal._get_raster_driver_name_from_path(out_path)
    driver = gdal.GetDriverByName(driver_name)

    if driver is None:
        raise ValueError(f"Unable to get driver for raster: {raster}")

    utils_path._delete_if_required(out_path, overwrite)

    if isinstance(dtype_str, str):
        dtype_str = dtype_str.lower()

    copy = driver.Create(
        out_path,
        metadata["width"],
        metadata["height"],
        metadata["bands"],
        utils_translate._translate_dtype_numpy_to_gdal(dtype_str),
        utils_gdal._get_default_creation_options(creation_options),
    )

    if copy is None:
        raise ValueError(f"Unable to create output raster: {out_path}")

    copy.SetProjection(metadata["projection_wkt"])
    copy.SetGeoTransform(metadata["geotransform"])

    for band_idx in range(metadata["bands"]):
        input_band = ref.GetRasterBand(band_idx + 1)
        output_band = copy.GetRasterBand(band_idx + 1)

        # Read the input band data and write it to the output band
        data = input_band.ReadAsArray(0, 0, input_band.XSize, input_band.YSize).astype(dtype_str)
        output_band.WriteRaster(0, 0, input_band.XSize, input_band.YSize, data)

        # Set the NoData value for the output band if it exists in the input band
        if input_band.GetNoDataValue() is not None:
            input_nodata = input_band.GetNoDataValue()
            if utils_translate._check_is_value_within_dtype_range(input_nodata, dtype_str):
                output_band.SetNoDataValue(input_nodata)
            else:
                warn("Input NoData value is outside the range of the output datatype. NoData value will not be set.", UserWarning)
                output_band.SetNoDataValue(None)

        # Set the color interpretation for the output band
        output_band.SetColorInterpretation(input_band.GetColorInterpretation())

    copy.FlushCache()

    ref = None
    copy = None

    return out_path


def raster_set_datatype(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    dtype: str,
    out_path: Optional[Union[str, List[str]]] = None,
    *,
    creation_options: Optional[List[str]] = None,
    add_uuid: bool = False,
    add_timestamp: bool = False,
    prefix: str = "",
    suffix: str = "",
    overwrite: bool = True,
) -> Union[str, List[str]]:
    """
    Converts the datatype of a raster.

    Parameters
    ----------
    raster : str or gdal.Dataset or list
        The input raster(s) for which the datatype will be changed.

    dtype : str
        The target datatype for the output raster(s).

    out_path : path or list, optional
        The output location for the processed raster(s). Default: None.

    creation_options : list, optional
        A list of GDAL creation options for the output raster(s). Default is
        ["TILED=YES", "NUM_THREADS=ALL_CPUS", "BIGTIFF=YES", "COMPRESS=LZW"].

    add_uuid : bool, optional
        Determines whether to add a UUID to the output path. Default: False.

    add_timestamp : bool, optional
        Determines whether to add a timestamp to the output path. Default: False.

    prefix : str, optional
        A prefix to add to the output path. Default: "".

    suffix : str, optional
        A suffix to add to the output path. Default: "".

    overwrite : bool, optional
        Determines whether to overwrite existing files with the same name. Default: True.

    Returns
    -------
    str or list
        The file path(s) of the newly created raster(s) with the specified datatype.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(dtype, [str, np.dtype, int, type(np.int8)], "dtype")
    utils_base._type_check(out_path, [list, str, None], "out_path")
    utils_base._type_check(creation_options, [list, None], "creation_options")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(overwrite, [bool], "overwrite")

    input_is_list = isinstance(raster, list)

    input_rasters = utils_io._get_input_paths(raster, "raster")
    out_paths = utils_io._get_output_paths(
        input_rasters,
        out_path,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        prefix=prefix,
        suffix=suffix,
    )

    creation_options = utils_gdal._get_default_creation_options(creation_options)

    utils_path._delete_if_required_list(out_paths, overwrite)

    output = []
    for idx, in_raster in enumerate(input_rasters):
        path = _raster_set_datatype(
            in_raster,
            dtype,
            out_path=out_paths[idx],
            overwrite=overwrite,
            creation_options=creation_options,
        )

        output.append(path)

    if input_is_list:
        return output

    return output[0]


def raster_get_datatype(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
) -> Union[str, List[str]]:
    """
    Gets the datatype of a raster.

    Parameters
    ----------
    raster : str or gdal.Dataset or list
        The input raster(s) for which the datatype will be changed.

    Returns
    -------
    str or list
        The datatype of the input raster(s).
    """
    input_is_list = isinstance(raster, list)
    input_rasters = utils_io._get_input_paths(raster, "raster")

    datatypes = []
    for input_raster in input_rasters:
        if not utils_gdal._check_is_raster(input_raster):
            raise ValueError(f"Unable to open input raster: {input_raster}")

        metadata = _get_basic_metadata_raster(input_raster)
        datatypes.append(metadata["dtype_name"])

    if input_is_list:
        return datatypes

    return datatypes[0]



"""
### Generic utility functions ###
Functions that make interacting with the toolbox easier.

"""
# Standard library
from typing import Union, Optional
import sys; sys.path.append("../../")

# External
import numpy as np
from numba import jit, prange

# Internal
from buteo.array.utils_array import _create_grid



# TODO: Multichannel support, Split assert
@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def convolve_fill_nearest(
    array: np.ndarray,
    nodata_value: Union[int, float],
    mask: Optional[np.ndarray] = None,
    max_iterations: Optional[Union[int, float]] = None,
    channel: int = 0,
):
    """
    Fill in nodata values with the average of the nearest values.
    
    Parameters
    ----------
    array : np.ndarray
        The array to fill in the nodata values for.

    nodata_value : Union[int, float]
        The value to use as the nodata value.

    mask : Optional[np.ndarray], optional
        The mask to use. Default: None.

    max_iterations : Optional[Union[int, float]], optional
        The maximum number of iterations to run. Default: None.

    channel : int, optional
        The channel to use. Default: 0.

    Returns
    -------
    np.ndarray
        The filled in array.
    """
    kernel_size = 3

    range_rows = np.arange(-(kernel_size // 2), (kernel_size // 2) + 1)
    range_cols = np.arange(-(kernel_size // 2), (kernel_size // 2) + 1)

    cols_grid, rows_grid = _create_grid(range_rows, range_cols)
    coord_grid = np.empty((cols_grid.size, 2), dtype=np.int64)
    coord_grid[:, 0] = cols_grid.flatten()
    coord_grid[:, 1] = rows_grid.flatten()

    coord_grid_values = np.sqrt((coord_grid[:, 0] ** 2) + (coord_grid[:, 1] ** 2))
    coord_grid_values_sort = np.argsort(coord_grid_values)[1:]
    coord_grid_values = coord_grid_values[coord_grid_values_sort]

    coord_grid = coord_grid[coord_grid_values_sort]

    weights = 1 / coord_grid_values
    weights = weights / np.sum(weights)
    weights = weights.astype(np.float32)

    main_filled = np.copy(array)

    if mask is None:
        mask = np.ones_like(main_filled, dtype=np.uint8)
    else:
        mask = (mask == 1).astype(np.uint8)

    main_filled = main_filled[:, :, channel]
    mask = mask[:, :, channel]

    nodata_value = np.array(nodata_value, dtype=array.dtype)
    uint8_1 = np.array(1, dtype=np.uint8)

    iterations = 0
    while True:
        local_filled = np.copy(main_filled)
        for row in prange(main_filled.shape[0]):
            for col in range(main_filled.shape[1]):
                if main_filled[row, col] != nodata_value:
                    continue
                if mask[row, col] != uint8_1:
                    continue

                count = 0
                weights_sum = 0.0
                value_sum = 0.0

                for idx, (col_adj, row_adj) in enumerate(coord_grid):
                    if (row + row_adj) >= 0 and (row + row_adj) < main_filled.shape[0] and \
                        (col + col_adj) >= 0 and (col + col_adj) < main_filled.shape[1] and \
                        main_filled[row + row_adj, col + col_adj] != nodata_value and \
                        mask[row + row_adj, col + col_adj] == uint8_1:

                        weight = weights[idx]
                        value = main_filled[row + row_adj, col + col_adj]

                        value_sum += value * weight
                        weights_sum += weight
                        count += 1

                if count == 0:
                    local_filled[row, col] = nodata_value
                else:
                    local_filled[row, col] = value_sum * (1.0 / weights_sum)

        main_filled = local_filled
        iterations += 1

        if max_iterations is not None and iterations >= max_iterations:
            break

        if np.sum((main_filled == nodata_value) & (mask == uint8_1)) == 0:
            break

    return np.expand_dims(main_filled, axis=2)



"""
### Handle and create borders on rasters. ###

Functions to add or remove borders from rasters.
Useful for warped satellite images and for proximity searching.
"""

# TODO: Remove near black borders.

# Standard library
import sys; sys.path.append("../../")
from typing import Union, List, Optional, Any

# External
import numpy as np
from osgeo import gdal

# Internal
from buteo.utils import (
    utils_io,
    utils_gdal,
    utils_base,
    utils_path,
    utils_translate,
)
from buteo.raster import core_raster, core_raster_io


def _raster_add_border(
    raster: Union[str, gdal.Dataset],
    out_path: Optional[str] = None,
    border_size: int = 100,
    *,
    border_size_unit: str = "px",
    border_value: Union[int, float] = 0,
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
):
    """
    Internal.
    Add a border to a raster.
    """
    in_raster = core_raster._raster_open(raster)
    metadata = core_raster._get_basic_metadata_raster(in_raster)

    # Parse the driver
    driver_name = "GTiff" if out_path is None else utils_gdal._get_raster_driver_name_from_path(out_path)
    if driver_name is None:
        raise ValueError(f"Unable to parse filetype from path: {out_path}")

    driver = gdal.GetDriverByName(driver_name)
    if driver is None:
        raise ValueError(f"Error while creating driver from extension: {out_path}")

    output_name = None
    if out_path is None:
        output_name = utils_path._get_temp_filepath("raster_proximity.tif", add_uuid=True, add_timestamp=True)
    else:
        assert utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite), (
            f"Output path is not valid: {out_path}"
        )
        output_name = out_path

    in_arr = core_raster_io.raster_to_array(in_raster)

    if border_size_unit == "px":
        border_size_y = border_size
        border_size_x = border_size
        new_shape = (
            in_arr.shape[0] + (2 * border_size_y),
            in_arr.shape[1] + (2 * border_size_x),
            in_arr.shape[2],
        )
    else:
        border_size_y = round(border_size / metadata["pixel_height"])
        border_size_x = round(border_size / metadata["pixel_width"])
        new_shape = (
            in_arr.shape[0] + (2 * border_size_y),
            in_arr.shape[1] + (2 * border_size_x),
            in_arr.shape[2],
        )

    new_arr = np.full(new_shape, border_value, dtype=in_arr.dtype)
    new_arr[border_size_y:-border_size_y, border_size_x:-border_size_x, :] = in_arr

    if isinstance(in_arr, np.ma.MaskedArray):
        mask = np.zeros(new_shape, dtype=bool)
        mask[
            border_size_y:-border_size_y, border_size_x:-border_size_x, :
        ] = in_arr.mask
        new_arr = np.ma.array(new_arr, mask=mask)
        new_arr.fill_value = in_arr.fill_value

    utils_path._delete_if_required(output_name, overwrite)

    dest_raster = driver.Create(
        output_name,
        new_shape[1],
        new_shape[0],
        metadata["bands"],
        utils_translate._translate_dtype_numpy_to_gdal(in_arr.dtype),
        utils_gdal._get_default_creation_options(creation_options),
    )

    og_transform = in_raster.GetGeoTransform()

    new_transform = []
    for i in og_transform:
        new_transform.append(i)

    new_transform[0] -= border_size_x * og_transform[1]
    new_transform[3] -= border_size_y * og_transform[5]

    dest_raster.SetGeoTransform(new_transform)
    dest_raster.SetProjection(in_raster.GetProjectionRef())

    for band_num in range(1, metadata["bands"] + 1):
        dst_band = dest_raster.GetRasterBand(band_num)
        dst_band.WriteArray(new_arr[:, :, band_num - 1])

        if metadata["nodata"]:
            dst_band.SetNoDataValue(metadata["nodata_value"])

    return output_name


def raster_add_border(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    out_path: Optional[str] = None,
    border_size: int = 100,
    border_size_unit: str = "px",
    border_value: int = 0,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
    ) -> Union[str, gdal.Dataset]:
    """
    Add a border to a raster.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]]
        The input raster.

    out_path : Optional[str], optional
        The output path. If None, the output will be a memory raster.
        Default: None

    border_size : int, optional
        The size of the border. Default: 100

    border_size_unit : str, optional
        The unit of the border size. Default: 'px'

    border_value : int, optional
        The value of the border. Default: 0

    prefix : str, optional
        The prefix to add to the output file name, default: ""

    suffix : str, optional
        The suffix to add to the output file name, default: ""

    add_uuid : bool, optional
        Whether to add a uuid to the output file name, default: False

    add_timestamp : bool, optional
        Whether to add a timestamp to the output file name, default: False

    overwrite : bool, optional
        If True, the output raster will be overwritten. Default: True

    creation_options : Optional[List[str]], optional
        Creation options for the output raster. Default is None

    Returns
    -------
    Union[str, gdal.Dataset]
        The output raster with added borders.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(out_path, [str, None], "out_path")
    utils_base._type_check(border_size, [int], "border_size")
    utils_base._type_check(border_size_unit, [str], "border_size_unit")
    utils_base._type_check(border_value, [int, float], "border_value")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(creation_options, [list, None], "creation_options")

    input_is_list = isinstance(raster, list)

    input_rasters = utils_io._get_input_paths(raster, "raster")
    out_path = utils_io._get_output_paths(
        input_rasters,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        change_ext="tif",
        overwrite=overwrite,
    )

    utils_path._delete_if_required_list(out_path, overwrite)

    for idx, raster in enumerate(input_rasters):
        _raster_add_border(
            raster,
            out_path=out_path[idx],
            border_size=border_size,
            border_size_unit=border_size_unit,
            border_value=border_value,
            overwrite=overwrite,
            creation_options=creation_options,
        )

    if input_is_list:
        return out_path

    return out_path[0]



"""
### Basic functionality for working with rasters. ###
"""

# Standard library
import sys; sys.path.append("../../")
import os
from typing import List, Optional, Union, Dict, Any
import warnings

# External
from osgeo import gdal, ogr, osr

# Internal
from buteo.utils import (
    utils_base,
    utils_gdal,
    utils_bbox,
    utils_path,
    utils_translate,
    utils_projection,
    utils_io,
)



def _raster_open(
    raster: Union[str, gdal.Dataset],
    *,
    writeable: bool = True,
) -> gdal.Dataset:
    """
    Opens a raster in read or write mode.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset]
        A path to a raster or a GDAL dataframe.

    writeable : bool, optional
        If True, the raster is opened in write mode. Default: True.
    
    Returns
    -------
    gdal.Dataset
        A gdal.Dataset.
    """
    assert isinstance(raster, (gdal.Dataset, str)), "raster must be a string or a gdal.Dataset"

    if isinstance(raster, gdal.Dataset):
        return raster

    if isinstance(raster, str) and raster.startswith("/vsizip/"):
        writeable = False

    if utils_path._check_file_exists(raster):

        gdal.PushErrorHandler("CPLQuietErrorHandler")
        opened = gdal.Open(raster, gdal.GF_Write) if writeable else gdal.Open(raster, gdal.GF_Read)
        gdal.PopErrorHandler()

        if not isinstance(opened, gdal.Dataset):
            raise ValueError(f"Input raster is not readable. Received: {raster}")

        if opened.GetDescription() == "":
            opened.SetDescription(raster)

        if opened.GetProjectionRef() == "":
            opened.SetProjection(utils_projection._get_default_projection())
            warnings.warn(f"WARNING: Input raster {raster} has no projection. Setting to default: EPSG:4326.", UserWarning)

        return opened

    raise ValueError(f"Input raster does not exists. Received: {raster}")


def _get_first_nodata_value(
    raster: Union[str, gdal.Dataset],
) -> Optional[Union[float, int]]:
    """
    Gets the first nodata value from a raster.

    Parameters
    ----------
    raster : str or gdal.Dataset
        The raster to get the nodata value from.

    Returns
    -------
    float or None
        The nodata value if found, or None if not found.
    """
    utils_base._type_check(raster, [str, gdal.Dataset], "raster")

    nodata = None

    dataset = _raster_open(raster)
    band_count = dataset.RasterCount
    for band in range(1, band_count + 1):
        band_ref = dataset.GetRasterBand(band)
        nodata_value = band_ref.GetNoDataValue()

        if nodata_value is not None:
            nodata = nodata_value
            break

    dataset = None
    return nodata


def _get_basic_metadata_raster(
    raster: Union[str, gdal.Dataset],
) -> Dict[str, Any]:
    """
    Get basic metadata from a raster.

    Parameters
    ----------
    raster : str or gdal.Dataset
        The raster to get the metadata from.

    Returns
    -------
    Dict[str]
        A dictionary with the metadata.
    """
    utils_base._type_check(raster, [str, gdal.Dataset], "raster")

    dataset = _raster_open(raster, writeable=False)
    transform = dataset.GetGeoTransform()
    projection_wkt = dataset.GetProjectionRef()
    projection_osr = osr.SpatialReference()
    projection_osr.ImportFromWkt(projection_wkt)

    bbox = utils_bbox._get_bbox_from_geotransform(transform, dataset.RasterXSize, dataset.RasterYSize)
    try:
        bbox_latlng = utils_projection.reproject_bbox(bbox, projection_osr, utils_projection._get_default_projection_osr())
    except RuntimeError as e: 
        # Catch domain errors which happen if the point requested is outside the projection domain
        # We set the latlng bbox to the default projection domain
        if e.args[0] == "Point outside of projection domain":
            bbox_latlng = [0.0, 90.0, 0.0, 180.0]
        else:
            raise e

    try:
        bounds_latlng = utils_bbox._get_bounds_from_bbox(bbox, projection_osr, wkt=False)
    except RuntimeError as e:
        if e.args[0] == "Point outside of projection domain":
            bounds_latlng = utils_bbox._get_bounds_from_bbox(bbox_latlng, utils_projection._get_default_projection_osr(), wkt=False)
        else:
            raise e

    bounds_area = bounds_latlng.GetArea()
    area = (bbox[1] - bbox[0]) * (bbox[3] - bbox[2])
    bounds_wkt = bounds_latlng.ExportToWkt()
    bounds_raster_raw = utils_bbox._get_geom_from_bbox(bbox)
    bounds_raster = bounds_raster_raw.ExportToWkt()
    first_band = dataset.GetRasterBand(1)
    dtype = None if first_band is None else first_band.DataType
    dtype_numpy = utils_translate._translate_dtype_gdal_to_numpy(dtype)

    _centroid = bounds_raster_raw.Centroid()
    centroid = (_centroid.GetX(), _centroid.GetY())
    if projection_osr.IsGeographic():
        centroid_latlng = (centroid[1], centroid[0])
    else:
        _centroid_latlng = bounds_latlng.Centroid()
        centroid_latlng = (_centroid_latlng.GetY(), _centroid_latlng.GetX())

    # Paths
    path = utils_path._get_unix_path(dataset.GetDescription())
    in_memory = utils_path._check_is_valid_mem_filepath(path)

    metadata = {
        "path": path,
        "basename": os.path.basename(path),
        "name": os.path.splitext(os.path.basename(path))[0],
        "folder": os.path.dirname(path),
        "ext": os.path.splitext(path)[1],
        "in_memory": in_memory,
        "driver": dataset.GetDriver().ShortName,
        "projection_osr": projection_osr,
        "projection_wkt": projection_wkt,
        "geotransform": transform,
        "size": (dataset.RasterXSize, dataset.RasterYSize),
        "shape": [dataset.RasterYSize, dataset.RasterXSize, dataset.RasterCount],
        "height": dataset.RasterYSize,
        "width": dataset.RasterXSize,
        "pixel_size": (abs(transform[1]), abs((transform[5]))),
        "pixel_width": abs(transform[1]),
        "pixel_height": abs(transform[5]),
        "origin": (transform[0], transform[3]),
        "origin_x": transform[0],
        "origin_y": transform[3],
        "centroid": centroid,
        "centroid_latlng": centroid_latlng,
        "bbox": bbox,
        "bbox_gdal": utils_bbox._get_gdal_bbox_from_ogr_bbox(bbox),
        "bbox_latlng": bbox_latlng,
        "bounds_latlng": bounds_wkt,
        "bounds_raster": bounds_raster,
        "x_min": bbox[0],
        "x_max": bbox[1],
        "y_min": bbox[2],
        "y_max": bbox[3],
        "bands": dataset.RasterCount,
        "dtype_gdal": dtype,
        "dtype": dtype_numpy,
        "dtype_name": dtype_numpy.name,
        "area_latlng": bounds_area,
        "area": area,
    }

    x_min, x_max, y_min, y_max = bbox
    metadata["area"] = (x_max - x_min) * (y_max - y_min)

    # Add the nodata values
    metadata["nodata"] = False
    metadata["nodata_value"] = None
    for band in range(1, metadata["bands"] + 1):
        band_ref = dataset.GetRasterBand(band)
        nodata_value = band_ref.GetNoDataValue()

        if nodata_value is not None:
            metadata["nodata"] = True
            metadata["nodata_value"] = nodata_value
            break

    dataset = None
    return metadata


def _get_basic_metadata_raster_list(
    rasters: List[Union[str, gdal.Dataset]],
) -> List[Dict[str, Any]]:
    """
    Get basic metadata from a list of rasters.

    Parameters
    ----------
    rasters : List[Union[str, gdal.Dataset]]
        The rasters to get the metadata from.

    Returns
    -------
    List[Dict[str]]
        A list of dictionaries with the metadata.
    """
    utils_base._type_check(rasters, [[str, gdal.Dataset]], "rasters")

    metadata = []
    for raster in rasters:
        metadata.append(_get_basic_metadata_raster(raster))

    return metadata


def _check_raster_has_nodata(
    raster: Union[str, gdal.Dataset],
) -> bool:
    """
    Verifies whether a raster has any nodata values.

    Parameters
    ----------
    raster : str or gdal.Dataset
        A raster, either in gdal.Dataset or a string referring to the dataset.

    Returns
    -------
    bool
        True if raster has nodata values, False otherwise.
    """
    utils_base._type_check(raster, [str, gdal.Dataset], "raster")

    metadata = _get_basic_metadata_raster(raster)
    if metadata["nodata"]:
        return True

    return False


def _check_raster_has_nodata_list(
    rasters: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
) -> bool:
    """
    Verifies whether a list of rasters have any nodata values.

    Parameters
    ----------
    rasters : list
        A list of rasters, either in gdal.Dataset or a string referring to the dataset.

    Returns
    -------
    bool
        True if all rasters have nodata values, False otherwise.
    """
    utils_base._type_check(rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")

    rasters = utils_io._get_input_paths(rasters, "raster")

    metadata = _get_basic_metadata_raster_list(rasters)
    for meta in metadata:
        if meta["nodata"]:
            return True

    return False


def _raster_count_bands_list(
    rasters: List[Union[str, gdal.Dataset]]
) -> int:
    """
    Counts the number of bands in a list of rasters.

    Parameters
    ----------
    rasters : list
        A list of rasters, either in gdal.Dataset or a string referring to the dataset.

    Returns
    -------
    int
        The number of bands in the rasters.
    """
    utils_base._type_check(rasters, [[str, gdal.Dataset]], "raster")

    rasters = utils_io._get_input_paths(rasters, "raster")
    metadata = _get_basic_metadata_raster_list(rasters)

    bands = 0
    for meta in metadata:
        bands += meta["bands"]

    return bands


def check_rasters_have_same_nodata(
    rasters: List[Union[str, gdal.Dataset]],
) -> bool:
    """
    Verifies whether a list of rasters have the same nodata values.

    Parameters
    ----------
    rasters : list
        A list of rasters, either in gdal.Dataset or a string referring to the dataset.

    Returns
    -------
    bool
        True if all rasters have the same nodata value, False otherwise.
    """
    utils_base._type_check(rasters, [[str, gdal.Dataset]], "raster")

    rasters = utils_io._get_input_paths(rasters, "raster")

    metadata = _get_basic_metadata_raster_list(rasters)

    nodata_value = metadata[0]["nodata_value"]
    for meta in metadata[1:]:
        if meta["nodata_value"] != nodata_value:
            return False

    return True


def check_rasters_intersect(
    raster1: Union[str, gdal.Dataset],
    raster2: Union[str, gdal.Dataset],
) -> bool:
    """
    Checks if two rasters intersect using their latlong boundaries.

    Parameters
    ----------
    raster1 : str or gdal.Dataset
        The first raster.

    raster2 : str or gdal.Dataset
        The second raster.

    Returns
    -------
    bool
        True if the rasters intersect, False otherwise.
    """
    utils_base._type_check(raster1, [str, gdal.Dataset], "raster1")
    utils_base._type_check(raster2, [str, gdal.Dataset], "raster2")

    meta_1 = _get_basic_metadata_raster(raster1)
    meta_2 = _get_basic_metadata_raster(raster2)

    geom_1 = ogr.CreateGeometryFromWkt(meta_1["bounds_latlng"], meta_1["projection_osr"])
    geom_2 = ogr.CreateGeometryFromWkt(meta_2["bounds_latlng"], meta_2["projection_osr"])

    # Do the layers intersect?
    intersect = geom_1.Intersects(geom_2)

    return intersect


def get_raster_intersection(
    raster1: Union[str, gdal.Dataset],
    raster2: Union[str, gdal.Dataset],
) -> ogr.Geometry:
    """
    Gets the latlng intersection of two rasters.

    Parameters
    ----------
    raster1 : str or gdal.Dataset
        The first raster.

    raster2 : str or gdal.Dataset
        The second raster.

    Returns
    -------
    tuple or ogr.Geometry
        If return_as_vector is False, returns a tuple `(xmin, ymin, xmax, ymax)` representing
        the intersection of the two rasters. If return_as_vector is True, returns an ogr.Geometry
        object representing the intersection.
    """
    utils_base._type_check(raster1, [str, gdal.Dataset], "raster1")
    utils_base._type_check(raster2, [str, gdal.Dataset], "raster2")

    if not check_rasters_intersect(raster1, raster2):
        raise ValueError("Rasters do not intersect.")

    meta_1 = _get_basic_metadata_raster(raster1)
    meta_2 = _get_basic_metadata_raster(raster2)

    geom_1 = ogr.CreateGeometryFromWkt(meta_1["bounds_latlng"], meta_1["projection_osr"])
    geom_2 = ogr.CreateGeometryFromWkt(meta_2["bounds_latlng"], meta_2["projection_osr"])

    if not geom_1.Intersects(geom_2):
        raise ValueError("Rasters do not intersect.")

    # Get the intersection.
    intersection = geom_1.Intersection(geom_2)

    return intersection


def get_raster_overlap_fraction(
    raster1: Union[str, gdal.Dataset],
    raster2: Union[str, gdal.Dataset],
) -> float:
    """
    Get the fraction of the overlap between two rasters.
    (e.g. 0.9 for mostly overlapping rasters)

    Parameters
    ----------
    raster1 : str or gdal.Dataset
        The first raster (master).

    raster2 : str or gdal.Dataset
        The second raster.

    Returns
    -------
    float
        A value between 0.0 and 1.0 representing the degree of overlap between the two rasters.
    """
    utils_base._type_check(raster1, [str, gdal.Dataset, [str, gdal.Dataset]], "raster1")
    utils_base._type_check(raster2, [str, gdal.Dataset, [str, gdal.Dataset]], "raster2")

    if not check_rasters_intersect(raster1, raster2):
        return 0.0

    meta_1 = _get_basic_metadata_raster(raster1)
    meta_2 = _get_basic_metadata_raster(raster2)

    geom_1 = ogr.CreateGeometryFromWkt(meta_1["bounds_latlng"], meta_1["projection_osr"])
    geom_2 = ogr.CreateGeometryFromWkt(meta_2["bounds_latlng"], meta_2["projection_osr"])

    if not geom_1.Intersects(geom_2):
        return 0.0

    try:
        intersection = geom_1.Intersection(geom_2)
        overlap = intersection.GetArea() / geom_1.GetArea()
    except RuntimeError:
        overlap = 0.0

    return overlap


def check_rasters_are_aligned(
    rasters: List[Union[str, gdal.Dataset]],
    *,
    same_dtype: bool = False,
    same_nodata: bool = False,
    same_bands: bool = False,
    threshold: float = 0.0001,
) -> bool:
    """
    Verifies whether a list of rasters are aligned.

    Parameters
    ----------
    rasters : list
        A list of rasters, either in gdal.Dataset or a string referring to the dataset.

    same_dtype : bool, optional
        If True, all the rasters should have the same data type. Default: False.

    same_nodata : bool, optional
        If True, all the rasters should have the same nodata value. Default: False.

    threshold : float, optional
        The threshold for the difference between the rasters. Default: 0.0001.

    Returns
    -------
    bool
        True if rasters are aligned and optional parameters are True, False otherwise.
    """
    utils_base._type_check(rasters, [[str, gdal.Dataset]], "rasters")
    utils_base._type_check(same_dtype, [bool], "same_dtype")
    utils_base._type_check(same_nodata, [bool], "same_nodata")
    utils_base._type_check(threshold, [float], "threshold")

    if len(rasters) == 0:
        raise ValueError("Input is an empty list.")

    if len(rasters) == 1:
        return True

    assert utils_gdal._check_is_raster_list(rasters), "Input is not a list of valid rasters."

    # Get the metadata of the first raster
    metadata_base = _get_basic_metadata_raster(rasters[0])

    for raster in rasters[1:]:
        # Get the metadata of the current raster
        metadata_current = _get_basic_metadata_raster(raster)

        # Check if the same amount of pixels
        if not metadata_base["size"][0] == metadata_current["size"][0] and metadata_base["size"][1] == metadata_current["size"][1]:
            return False

        # Check if the projections are the same
        if not metadata_base["projection_osr"].IsSame(metadata_current["projection_osr"]):
            return False

        # Check if origin is the same
        origin_base = metadata_base["origin"]
        origin_current = metadata_current["origin"]

        if not utils_base._check_number_is_within_threshold(origin_base[0], origin_current[0], threshold):
            return False

        if not utils_base._check_number_is_within_threshold(origin_base[1], origin_current[1], threshold):
            return False

        # Check if the pixel size is the same
        pixel_size_base = metadata_base["pixel_size"]
        pixel_size_current = metadata_current["pixel_size"]

        if not utils_base._check_number_is_within_threshold(pixel_size_base[0], pixel_size_current[0], threshold):
            return False

        if not utils_base._check_number_is_within_threshold(pixel_size_base[1], pixel_size_current[1], threshold):
            return False

        if same_dtype:
            if not metadata_base["dtype"] == metadata_current["dtype"]:
                return False

        if same_nodata:
            if not metadata_base["nodata"] == metadata_current["nodata"]:
                return False

            if metadata_base["nodata"]:
                if not metadata_base["nodata_value"] == metadata_current["nodata_value"]:
                    return False

        if same_bands:
            if not metadata_base["bands"] == metadata_current["bands"]:
                return False

    return True


def raster_to_extent(
    raster: Union[str, gdal.Dataset],
    out_path: Optional[str] = None,
    *,
    latlng: bool = False,
    overwrite: bool = True,
) -> str:
    """
    Converts a raster to a vector file with the extent as a polygon.

    Parameters
    ----------
    raster : str or gdal.Dataset
        The raster to convert.

    out_path : str, optional
        The path to save the extent to. If None, the extent is saved in memory. Default: None.

    latlng : bool, optional
        If True, the extent is returned in latlng coordinates. If false,
        the projection of the raster is used. Default: False.

    overwrite : bool, optional
        If True, the output file is overwritten if it exists. Default: True.

    Returns
    -------
    str
        The path to the extent.
    """
    utils_base._type_check(raster, [str, gdal.Dataset], "raster")
    utils_base._type_check(out_path, [str, type(None)], "out_path")
    utils_base._type_check(latlng, [bool], "latlng")
    utils_base._type_check(overwrite, [bool], "overwrite")

    if out_path is None:
        out_path = utils_path._get_temp_filepath("temp_extent.gpkg", add_timestamp=True, add_uuid=True)

    if not utils_path._check_is_valid_output_filepath(out_path):
        raise ValueError(f"Invalid output path: {out_path}")

    metadata = _get_basic_metadata_raster(raster)

    if latlng:
        extent = metadata["bounds_latlng"]
    else:
        extent = metadata["bounds_raster"]

    extent = ogr.CreateGeometryFromWkt(extent, metadata["projection_osr"])

    driver_name = utils_gdal._get_driver_name_from_path(out_path)

    driver = ogr.GetDriverByName(driver_name)
    extent_ds = driver.CreateDataSource(out_path)
    extent_layer = extent_ds.CreateLayer("extent", metadata["projection_osr"], ogr.wkbPolygon)
    extent_feature = ogr.Feature(extent_layer.GetLayerDefn())
    extent_feature.SetGeometry(extent)
    extent_layer.CreateFeature(extent_feature)

    extent_ds = None
    extent_layer = None
    extent_feature = None

    return out_path


def raster_open(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    *,
    writeable=True,
) -> Union[gdal.Dataset, List[gdal.Dataset]]:
    """
    Opens a raster in read or write mode.
    Supports lists.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]]
        A path to a raster or a GDAL dataframe.

    writeable : bool, optional
        If True, the raster is opened in write mode. Default: True.

    Returns
    -------
    Union[gdal.Dataset, List[gdal.Dataset]]
        A gdal.Dataset or a list of gdal.Datasets.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(writeable, [bool], "writeable")

    input_is_list = isinstance(raster, list)
    in_raster = utils_io._get_input_paths(raster, "raster")

    list_return = []
    for r in in_raster:
        try:
            list_return.append(_raster_open(r, writeable=writeable))
        except Exception:
            raise ValueError(f"Could not open raster: {r}") from None

    if input_is_list:
        return list_return

    return list_return[0]



"""
### Handle nodata values in rasters. ###

A module to handle the various aspects of NODATA in raster files.
"""

# TODO: raster_to_mask
# TODO: raster_invert_nodata

# Standard library
import sys; sys.path.append("../../")
from typing import List, Union, Optional

# External
from osgeo import gdal

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_gdal,
    utils_path,
    utils_translate,
)
from buteo.raster import core_raster, core_raster_io



def _raster_has_nodata(
    raster: Union[str, gdal.Dataset],
) -> bool:
    """
    Internal. Check if a raster or a list of rasters contain nodata values.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset]
        The raster to check for nodata values.

    Returns
    -------
    bool
        True if the raster has nodata values
    """
    assert isinstance(raster, (str, gdal.Dataset)), f"Invalid raster. {raster}"

    metadata = core_raster._get_basic_metadata_raster(raster)

    if metadata["nodata"]:
        return True

    return False


def raster_has_nodata(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
) -> Union[bool, List[bool]]:
    """
    Check if a raster or a list of rasters contain nodata values.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset, List]
        The raster to check for nodata values.

    Returns
    -------
    Union[bool, List[bool]
        True if the raster or list of rasters contain nodata values.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")

    input_is_list = isinstance(raster, list)

    input_paths = utils_io._get_input_paths(raster, "raster")

    nodata_values = []
    for in_raster in input_paths:
        nodata_values.append(_raster_has_nodata(in_raster))

    if input_is_list:
        return nodata_values

    return nodata_values[0]


def _raster_get_nodata(
    raster: Union[str, gdal.Dataset],
) -> Union[float, int]:
    """
    Internal. Get the nodata value of a raster.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset]
        The raster to get nodata values from.

    Returns
    -------
    Union[float, int]
        The nodata value of the raster.
    """
    utils_base._type_check(raster, [str, gdal.Dataset], "raster")

    metadata = core_raster._get_basic_metadata_raster(raster)

    return metadata["nodata_value"]


def raster_get_nodata(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
) -> Union[float, int, List[Union[float, int]]]:
    """
    Get the nodata value of a raster or a list of rasters.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset, List]
        The raster to get nodata values from.

    Returns
    -------
    Union[float, int, None, List[Union[float, int, None]]]
        The nodata value of the raster or list of rasters.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")

    input_is_list = isinstance(raster, list)

    input_paths = utils_io._get_input_paths(raster, "raster")

    nodata_values = []
    for in_raster in input_paths:
        nodata_values.append(_raster_get_nodata(in_raster))

    if input_is_list:
        return nodata_values

    return nodata_values[0]


def _raster_set_nodata(
    raster: Union[str, gdal.Dataset],
    nodata: Union[float, int, None],
    out_path: Optional[str] = None,
    in_place: bool = True,
    overwrite: bool = True,
) -> Union[str, gdal.Dataset]:
    """
    Internal. Sets the nodata value of a single raster.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset]
        The raster to set nodata values for.
    
    nodata : float, int, or None
        The nodata value to set for the raster.
        If nodata is None, nodata will be removed from the raster.

    out_path : str, optional
        The destination of the changed raster, default: None

    in_place : bool, optional
        Should the raster be changed in_place or copied?, default: True

    Returns
    -------
    str
        Returns the path to the raster with nodata set.
    """
    assert isinstance(raster, (str, gdal.Dataset)), f"Invalid raster. {raster}"
    assert isinstance(nodata, (float, int, type(None))), f"Invalid nodata value. {nodata}"
    assert isinstance(out_path, (str, type(None))), f"Invalid out_path. {out_path}"
    assert isinstance(in_place, bool), f"Invalid in_place. {in_place}"
    assert isinstance(overwrite, bool), f"Invalid overwrite. {overwrite}"

    if in_place:
        opened = core_raster._raster_open(raster)
    else:
        copy = core_raster_io.raster_create_copy(raster, out_path, overwrite=overwrite)
        opened = core_raster._raster_open(copy)

    bands = opened.RasterCount

    dtype = utils_translate._translate_dtype_gdal_to_numpy(opened.GetRasterBand(1).DataType)

    if nodata is not None:
        if not utils_translate._check_is_value_within_dtype_range(nodata, dtype):
            raise ValueError(f"Invalid nodata value for datatype. value: {nodata}, dtype: {dtype}")

    if nodata is None:
        for band in range(bands):
            raster_band = opened.GetRasterBand(band + 1)
            raster_band.DeleteNoDataValue()
            raster_band = None
    else:
        for band in range(bands):
            raster_band = opened.GetRasterBand(band + 1)
            raster_band.SetNoDataValue(nodata)
            raster_band = None

    if in_place:
        out_path = utils_gdal._get_path_from_dataset(opened)

    opened.FlushCache()
    opened = None

    return out_path


def raster_set_nodata(
    raster: Union[str, gdal.Dataset, List],
    nodata: Union[float, int, None],
    out_path: Optional[str] = None,
    in_place: bool = True,
    *,
    overwrite: bool = True,
    prefix: str = "",
    suffix: str = "",
    creation_options: Optional[List] = None,
):
    """
    Sets all the nodata for raster(s) to a value.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset, List]
        The raster(s) to set nodata values for.

    nodata : float, int, or None
        The nodata value to set for the raster(s).
        If nodata is None, nodata will be removed from the raster(s).
    
    out_path : str, optional
        The destination of the changed rasters, default: None

    in_place : bool, optional
        Should the rasters be changed in_place or copied?, default: True

    overwrite : bool, optional
        Should the rasters be overwritten if they already exist? default: True

    prefix : str, optional
        Prefix to add to the output, default: ""

    suffix : str, optional
        Suffix to add to the output, default: "_nodata_set"

    creation_options : List, optional
        Creation options for the output rasters, default: None

    Returns
    -------
    Union[str, List]
        Returns the rasters with nodata set.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(nodata, [float, int, None], "nodata")
    utils_base._type_check(out_path, [list, str, None], "out_path")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "postfix")
    utils_base._type_check(creation_options, [[str], None], "creation_options")

    input_is_list = isinstance(raster, list)

    input_paths = utils_io._get_input_paths(raster, "raster")
    output_paths = utils_io._get_output_paths(
        input_paths,
        out_path,
        in_place=in_place,
        prefix=prefix,
        suffix=suffix,
    )

    if not in_place:
        utils_path._delete_if_required_list(output_paths, overwrite)

    nodata_set = []
    for idx, in_raster in enumerate(input_paths):
        nodata_set.append(_raster_set_nodata(
            in_raster,
            nodata,
            out_path=output_paths[idx],
            in_place=in_place,
            overwrite=overwrite,
        ))

    if input_is_list:
        return output_paths

    return output_paths[0]
